# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import re
from queue import PriorityQueue

from . import categorytype


class CharacterCategory(object):

    class Range(object):

        def __lt__(self, other):
            return self.high < other.high

        def __init__(self, low=0, high=0, categories=None):
            self.low = low
            self.high = high
            self.categories = categories or []

        def contains(self, cp):
            return self.low <= cp < self.high

        def containing_length(self, text):
            for i in range(len(text)):
                c = ord(text[i])
                if c < self.low or c > self.high:
                    return i
            return len(text)

        def lower(self, cp):
            return self.high <= cp

        def higher(self, cp):
            return self.low > cp

        def match(self, other):
            return self.low == other.low and self.high == other.high

    def __init__(self):
        self.range_list = []

    def _compile(self) -> None:
        """
        _compile transforms self.range_list to non overlapped range list
        to apply binary search in get_category_types
        :return:
        """
        self.range_list.sort(key=lambda x: x.high)
        self.range_list.sort(key=lambda x: x.low)
        new_range_list = []
        left_chain = PriorityQueue()
        right_chain = self.range_list
        states = []
        pivot = 0
        while True:
            if left_chain.empty():
                if not right_chain:
                    break
                right = right_chain.pop(0)
                left_chain.put(right)
                pivot = right.low
                states.extend(right.categories)
                continue
            left = left_chain.get()
            right = right_chain[0] if right_chain else None
            left_end = left.high
            right_begin = right.low if right else math.inf
            if left_end <= right_begin:
                new_range_list.append(self.Range(pivot, left_end, set(states)))
                pivot = left_end
                for cat in left.categories:
                    states.remove(cat)
                continue
            else:
                new_range_list.append(self.Range(pivot, right_begin, set(states)))
                pivot = right_begin
                states.extend(right.categories)
                left_chain.put(right)
                left_chain.put(left)
                right_chain.pop(0)
        self.range_list = []
        _range = new_range_list[0]
        for irange in new_range_list[1:]:
            if irange.low == _range.high and irange.categories == _range.categories:
                _range = self.Range(_range.low, irange.high, _range.categories)
            else:
                self.range_list.append(_range)
                _range = irange
        self.range_list.append(_range)

    def get_category_types(self, code_point):
        begin = 0
        n = len(self.range_list)
        end = n
        pivot = (begin + end) // 2
        while 0 <= pivot < n:
            range_ = self.range_list[pivot]
            if range_.contains(code_point):
                return range_.categories
            if range_.lower(code_point):
                begin = pivot
            else:  # range_.higher(code_point)
                end = pivot
            new_pivot = (begin + end) // 2
            if new_pivot == pivot:
                break
            pivot = new_pivot
        return {categorytype.CategoryType.DEFAULT}

    def read_character_definition(self, char_def=None):
        """
        :param char_def: path
        """

        if char_def is not None:
            f = open(char_def, 'r', encoding="utf-8")
        else:
            f = open("char.def", 'r', encoding="utf-8")

        for i, line in enumerate(f.readlines()):
            line = line.rstrip()
            if re.fullmatch(r"\s*", line) or re.match("#", line):
                continue
            cols = re.split(r"\s+", line)
            if len(cols) < 2:
                f.close()
                raise AttributeError("invalid format at line {}".format(i))
            if not re.match("0x", cols[0]):
                continue
            range_ = self.Range()
            r = re.split("\\.\\.", cols[0])
            range_.low = int(r[0], 16)
            range_.high = range_.low + 1
            if len(r) > 1:
                range_.high = int(r[1], 16) + 1
            if range_.low >= range_.high:
                f.close()
                raise AttributeError("invalid range at line {}".format(i))
            for j in range(1, len(cols)):
                if re.match("#", cols[j]) or cols[j] == '':
                    break
                type_ = categorytype.CategoryType.get(cols[j])
                if type_ is None:
                    f.close()
                    raise AttributeError("{} is invalid type at line {}".format(cols[j], i))
                range_.categories.append(type_)
            self.range_list.append(range_)

        f.close()
        self._compile()



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class CategoryType(Enum):
    DEFAULT = 1
    SPACE = 1 << 1
    KANJI = 1 << 2
    SYMBOL = 1 << 3
    NUMERIC = 1 << 4
    ALPHA = 1 << 5
    HIRAGANA = 1 << 6
    KATAKANA = 1 << 7
    KANJINUMERIC = 1 << 8
    GREEK = 1 << 9
    CYRILLIC = 1 << 10
    USER1 = 1 << 11
    USER2 = 1 << 12
    USER3 = 1 << 13
    USER4 = 1 << 14
    NOOOVBOW = 1 << 15

    def get_id(self):
        return self.id

    def get_type(self, id_):
        for type_ in CategoryType.values():
            if type_.get_id() is id_:
                return type_
        return None

    @staticmethod
    def get(str_):
        try:
            return CategoryType[str_]
        except KeyError:
            return None



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from enum import Enum
from typing import List

from .dictionarylib.categorytype import CategoryType
from .dictionarylib.grammar import Grammar
from .dictionarylib.lexicon import Lexicon
from .lattice cimport Lattice
from .latticenode cimport LatticeNode
from .morphemelist import MorphemeList
from .plugin.input_text import InputTextPlugin
from .plugin.path_rewrite import PathRewritePlugin
from .utf8inputtext import UTF8InputText
from .utf8inputtextbuilder import UTF8InputTextBuilder


cdef void build_lattice_c(object tokenizer, object input_):
    bytes_ = input_.get_byte_text()

    cdef Lattice lattice = tokenizer._lattice
    lattice.resize(len(bytes_))

    cdef unsigned int i, word_id, end, idx
    cdef int left_id, right_id, cost
    cdef object lexicon = tokenizer._lexicon
    cdef list oov_provider_plugins = tokenizer._oov_provider_plugins

    for i in range(len(bytes_)):
        if not input_.can_bow(i) or not lattice.has_previous_node(i):
            continue
        iterator = lexicon.lookup(bytes_, i)
        has_words = False
        for word_id, end in iterator:
            if (end < len(bytes_)) and (not input_.can_bow(end)):
                continue
            has_words = True

            lex = lexicon.lexicons[word_id >> 28]
            idx = (0x0FFFFFFF & word_id) * 3 # 3 is ELEMENT_SIZE_AS_SHORT
            left_id, right_id, cost = lex.word_params._array_view[idx:idx+3]
            n = LatticeNode(lexicon, left_id, right_id, cost, word_id)

            lattice.insert(i, end, n)

        # OOV
        if CategoryType.NOOOVBOW not in input_.get_char_category_types(i):
            for oov_plugin in tokenizer._oov_provider_plugins:
                for node in oov_plugin.get_oov(input_, i, has_words):
                    has_words = True
                    lattice.insert(node.get_begin(), node.get_end(), node)
        if not has_words and tokenizer.default_oov_provider:
            for node in tokenizer.default_oov_provider.get_oov(input_, i, has_words):
                has_words = True
                lattice.insert(node.get_begin(), node.get_end(), node)

        if not has_words:
            raise RuntimeError("there is no morpheme at " + str(i))
    lattice.connect_node(lattice.eos_node)

class Tokenizer:
    """ tokenizer of morphological analysis

    Attributes:
        SplitMode:
            split mode to change words continuation.
            A == short mode
            B == middle mode
            C == long mode
        _dump_output:
            file object to dump lattice structure
        _grammar:

        _input_text_plugins:

        _lattice:

        _lexicon:

        _mode:

        _oov_provider_plugins:

        _path_rewrite_plugins:

    """

    SplitMode = Enum("SplitMode", "A B C")

    def __init__(self, grammar: Grammar, lexicon: Lexicon, input_text_plugins: List[InputTextPlugin],
                 oov_provider_plugins: List, path_rewrite_plugins: List[PathRewritePlugin],
                 mode: SplitMode = None):
        self._grammar = grammar
        self._lexicon = lexicon
        self._input_text_plugins = input_text_plugins
        self._oov_provider_plugins = oov_provider_plugins
        self._path_rewrite_plugins = path_rewrite_plugins
        self._dump_output = open(os.devnull, 'w')
        self._lattice = Lattice(grammar)
        self._mode = mode or self.SplitMode.C
        self._logger = logging.getLogger(__name__)
        self._logger.disabled = True
        if self._oov_provider_plugins:
            self.default_oov_provider = self._oov_provider_plugins[-1]

    def tokenize(self, text: str, mode=None, logger=None) -> MorphemeList:
        """ tokenize a text.

        In default tokenize text with SplitMode.C

        Args:
            text: input text
            mode: split mode
            logger: if True output lattice structure
        Returns:
            list of morpheme (MorphemeList)

        """
        if not text:
            return MorphemeList.empty()

        mode = mode or self._mode
        logger = logger or self._logger

        builder = UTF8InputTextBuilder(text, self._grammar)
        for plugin in self._input_text_plugins:
            plugin.rewrite(builder)
        input_ = builder.build()
        logger.info('=== Inupt dump:')
        logger.info(input_.get_text())

        self._build_lattice(input_)

        logger.info('=== Lattice dump:')
        self._lattice.dump(logger)

        path = self._lattice.get_best_path()

        logger.info('=== Before Rewriting:')
        self._dump_path(path, logger)

        for plugin in self._path_rewrite_plugins:
            plugin.rewrite(input_, path, self._lattice)
        self._lattice.clear()

        path = self._split_path(path, mode)

        logger.info('=== After Rewriting:')
        self._dump_path(path, logger)
        logger.info('===')

        ml = MorphemeList(input_, self._grammar, self._lexicon, path)
        return ml

    def _build_lattice(self, input_: UTF8InputText):
        build_lattice_c(self, input_)

    def _split_path(self, path: List[LatticeNode], mode: SplitMode) -> List[LatticeNode]:
        if mode == self.SplitMode.C:
            return path
        new_path = []
        for node in path:
            if mode is self.SplitMode.A:
                wids = node.get_word_info().a_unit_split
            else:
                wids = node.get_word_info().b_unit_split
            if len(wids) <= 1:
                new_path.append(node)
            else:
                offset = node.get_begin()
                for wid in wids:
                    n = LatticeNode(self._lexicon, 0, 0, 0, wid)
                    n.set_begin(offset)
                    offset += n.get_word_info().head_word_length
                    n.set_end(offset)
                    new_path.append(n)
        return new_path

    def _dump_path(self, path: List[LatticeNode], logger) -> None:
        if logger.disabled:
            return
        for i, node in enumerate(path):
            logger.info('{}: {}'.format(i, node))



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mmap

from .dictionaryversion import has_grammar, has_synonym_group_ids, is_dictionary
from .dictionaryheader import DictionaryHeader
from .doublearraylexicon import DoubleArrayLexicon
from .grammar import Grammar


class BinaryDictionary(object):

    def __init__(self, bytes_: mmap.mmap, grammar: Grammar, header: DictionaryHeader, lexicon: DoubleArrayLexicon):
        self._bytes = bytes_
        self._grammar = grammar
        self._header = header
        self._lexicon = lexicon

    @staticmethod
    def _read_dictionary(filename, access=mmap.ACCESS_READ):
        with open(filename, 'rb') as system_dic:
            bytes_ = mmap.mmap(system_dic.fileno(), 0, access=access)
        offset = 0
        header = DictionaryHeader.from_bytes(bytes_, offset)
        offset += header.storage_size()
        if not is_dictionary(header.version):
            raise Exception('invalid dictionary version')
        grammar = None
        if has_grammar(header.version):
            grammar = Grammar(bytes_, offset)
            offset += grammar.get_storage_size()

        lexicon = DoubleArrayLexicon(bytes_, offset, has_synonym_group_ids(header.version))
        return bytes_, grammar, header, lexicon

    @classmethod
    def from_system_dictionary(cls, filename):
        args = cls._read_dictionary(filename)
        if not args[2].is_system_dictionary():
            raise IOError('invalid system dictionary')
        return cls(*args)

    @classmethod
    def from_user_dictionary(cls, filename):
        args = cls._read_dictionary(filename, mmap.ACCESS_COPY)
        if not args[2].is_user_dictionary():
            raise IOError('invalid user dictionary')
        return cls(*args)

    def close(self):
        del self._grammar
        del self._lexicon
        self._bytes.close()

    @property
    def grammar(self) -> Grammar:
        return self._grammar

    @property
    def header(self) -> DictionaryHeader:
        return self._header

    @property
    def lexicon(self) -> DoubleArrayLexicon:
        return self._lexicon



cdef class LatticeNode:

    cdef int begin
    cdef int end
    cdef int total_cost
    cdef int word_id
    cdef bint _is_oov
    cdef LatticeNode best_previous_node
    cdef bint is_connected_to_bos
    cdef object extra_word_info
    cdef object undefined_word_info
    cdef bint _is_defined
    cdef object lexicon
    cdef int left_id
    cdef int right_id
    cdef int cost




# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

from sudachipy.dictionarylib.grammar import Grammar


class EditConnectionCostPlugin(ABC):
    """ Interface of plugins for editing the connection costs.

    The following is an example of settings.
    ``
    {
        {
            "class" : "sudachipy.plugin.connect_cost.SampleEditConnectionPlugin",
            "example" : "example setting"
        }
    }
    ``

    """

    @abstractmethod
    def set_up(self, grammar: Grammar) -> None:
        """ Set up plugin.

        Args:
            grammar: grammar of system dictionary

        """
        raise NotImplementedError

    @abstractmethod
    def edit(self, grammar: Grammar) -> None:
        """ Edit a connection costs.

        Args:
            grammar: grammar of system dictionary

        """
        raise NotImplementedError

    @staticmethod
    def inhibit_connection(grammar: Grammar, left: int, right: int) -> None:
        """ Inhibit a connection.

        Args:
            grammar: grammar of system dictionary
            left: right-ID of left node
            right: left-ID of right node

        """
        grammar.set_connect_cost(left, right, Grammar.INHIBITED_CONNECTION)



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import re
from logging import DEBUG, StreamHandler, getLogger

from dartsclone import DoubleArray

from sortedcontainers import SortedDict

from sudachipy.dictionarylib.jtypedbytebuffer import JTypedByteBuffer
from sudachipy.dictionarylib.wordinfo import WordInfo


class DictionaryBuilder(object):

    __BYTE_MAX_VALUE = 127
    __MAX_LENGTH = 255
    __MIN_REQUIRED_COLS_NUM = 18
    __BUFFER_SIZE = 1024 * 1024
    __PATTERN_UNICODE_LITERAL = re.compile(r"\\u([0-9a-fA-F]{4}|{[0-9a-fA-F]+})")
    __ARRAY_MAX_LENGTH = __BYTE_MAX_VALUE  # max value of byte in Java
    __STRING_MAX_LENGTH = 32767  # max value of short in Java
    is_user_dictionary = False

    class WordEntry:
        headword = None
        parameters = None
        wordinfo = None
        aunit_split_string = None
        bunit_split_string = None
        cunit_split_string = None

    class PosTable(object):

        def __init__(self):
            self.table = []

        def get_id(self, str_):
            id_ = self.table.index(str_) if str_ in self.table else -1
            if id_ < 0:
                id_ = len(self.table)
                self.table.append(str_)
            return id_

        def get_list(self):
            return self.table

    @staticmethod
    def __default_logger():
        handler = StreamHandler()
        handler.terminator = ""
        handler.setLevel(DEBUG)
        logger = getLogger(__name__)
        logger.setLevel(DEBUG)
        logger.addHandler(handler)
        logger.propagate = False
        return logger

    def __init__(self, *, logger=None):
        self.byte_buffer = JTypedByteBuffer()
        self.trie_keys = SortedDict()
        self.entries = []
        self.is_dictionary = False
        self.pos_table = self.PosTable()
        self.logger = logger or self.__default_logger()

    def build(self, lexicon_paths, matrix_input_stream, out_stream):
        self.logger.info('reading the source file...')
        for path in lexicon_paths:
            with open(path, 'r', encoding='utf-8') as rf:
                self.build_lexicon(rf)
        self.logger.info('{} words\n'.format(len(self.entries)))

        self.write_grammar(matrix_input_stream, out_stream)
        self.write_lexicon(out_stream)

    def build_lexicon(self, lexicon_input_stream):
        line_no = -1
        try:
            for i, row in enumerate(csv.reader(lexicon_input_stream)):
                line_no = -1
                entry = self.parse_line(row)
                if entry.headword:
                    self.add_to_trie(entry.headword, len(self.entries))
                self.entries.append(entry)
        except Exception as e:
            if line_no > 0:
                self.logger.error(
                    '{} at line {} in {}\n'.format(e.args[0], line_no, lexicon_input_stream.name))
            raise e

    def parse_line(self, cols):
        if len(cols) < self.__MIN_REQUIRED_COLS_NUM:
            raise ValueError('invalid format')
        cols = [self.decode(col) for col in cols]
        if not self.__is_length_valid(cols):
            raise ValueError('string is too long')
        if not cols[0]:
            raise ValueError('headword is empty')

        entry = self.WordEntry()
        # head word for trie
        if cols[1] != '-1':
            entry.headword = cols[0]
        # left-id, right-id, connect_cost
        entry.parameters = [int(cols[i]) for i in [1, 2, 3]]
        # part of speech
        pos_id = self.get_posid(cols[5:11])
        if pos_id < 0:
            raise ValueError('invalid part of speech')

        entry.aunit_split_string = cols[15]
        entry.bunit_split_string = cols[16]
        entry.cunit_split_string = cols[17]
        self.check_splitinfo_format(entry.aunit_split_string)
        self.check_splitinfo_format(entry.bunit_split_string)
        self.check_splitinfo_format(entry.cunit_split_string)

        if cols[14] == 'A' and \
                not (entry.aunit_split_string == '*' and entry.bunit_split_string == '*'):
            raise ValueError('invalid splitting')

        synonym_group_ids = []
        if len(cols) > 18:
            synonym_group_ids = self.parse_synonym_group_ids(cols[18])

        head_length = len(cols[0].encode('utf-8'))
        dict_from_wordid = -1 if cols[13] == '*' else int(cols[13])
        entry.wordinfo = WordInfo(
            cols[4], head_length, pos_id, cols[12], dict_from_wordid, '', cols[11], None, None, None, synonym_group_ids)
        return entry

    def __is_length_valid(self, cols):
        head_length = len(cols[0].encode('utf-8'))
        return head_length <= self.__STRING_MAX_LENGTH \
            and len(cols[4]) <= self.__STRING_MAX_LENGTH \
            and len(cols[11]) <= self.__STRING_MAX_LENGTH \
            and len(cols[12]) <= self.__STRING_MAX_LENGTH

    def add_to_trie(self, headword, word_id):
        key = headword.encode('utf-8')
        if key not in self.trie_keys:
            self.trie_keys[key] = []
        self.trie_keys[key].append(word_id)

    def get_posid(self, strs):
        return self.pos_table.get_id(','.join(strs))

    def write_grammar(self, matrix_input_stream, output_stream):
        self.logger.info('writing the POS table...')
        self.convert_postable(self.pos_table.get_list())
        self.byte_buffer.seek(0)
        output_stream.write(self.byte_buffer.read())
        self.__logging_size(self.byte_buffer.tell())
        self.byte_buffer.clear()

        self.logger.info('writing the connection matrix...')
        if not matrix_input_stream:
            self.byte_buffer.write_int(0, 'short')
            self.byte_buffer.write_int(0, 'short')
            self.byte_buffer.seek(0)
            output_stream.write(self.byte_buffer.read())
            self.__logging_size(self.byte_buffer.tell())
            self.byte_buffer.clear()
            return
        matrix = self.convert_matrix(matrix_input_stream)
        self.byte_buffer.seek(0)
        output_stream.write(self.byte_buffer.read())
        self.byte_buffer.clear()
        output_stream.write(matrix.read())
        self.__logging_size(matrix.tell() + 4)

    def convert_postable(self, pos_list):
        self.byte_buffer.write_int(len(pos_list), 'short')
        for pos in pos_list:
            for text in pos.split(','):
                self.write_string(text)

    def convert_matrix(self, matrix_input):
        header = matrix_input.readline().strip()
        if re.fullmatch(r"\s*", header):
            raise ValueError('invalid format at line 0')
        lr = header.split()
        lsize, rsize = [int(x) for x in lr]
        self.byte_buffer.write_int(lsize, 'short')
        self.byte_buffer.write_int(rsize, 'short')

        matrix = JTypedByteBuffer()

        for i, line in enumerate(matrix_input.readlines()):
            line = line.strip()
            if re.fullmatch(r"\s*", line) or re.match("#", line):
                continue
            cols = line.split()
            if len(cols) < 3:
                self.logger.warn('invalid format at line {}'.format(i))
                continue
            l, r, cost = [int(col) for col in cols]
            pos = matrix.tell()
            matrix.seek(2 * (l + lsize * r))
            matrix.write_int(cost, 'short')
            matrix.seek(pos)
        return matrix

    def write_lexicon(self, io_out):
        trie = DoubleArray()
        wordid_table = JTypedByteBuffer()
        keys = []
        vals = []
        for key, word_ids in self.trie_keys.items():
            keys.append(key)
            vals.append(wordid_table.tell())
            wordid_table.write_int(len(word_ids), 'byte')
            for wid in word_ids:
                wordid_table.write_int(wid, 'int')

        self.logger.info('building the trie...')

        trie.build(keys, lengths=[len(k) for k in keys], values=vals)

        self.logger.info('done\n')
        self.logger.info('writing the trie...')
        self.byte_buffer.clear()
        self.byte_buffer.write_int(trie.size(), 'int')
        self.byte_buffer.seek(0)
        io_out.write(self.byte_buffer.read())
        self.byte_buffer.clear()

        io_out.write(trie.array())
        self.__logging_size(trie.size() * 4 + 4)
        trie.clear()
        del trie

        self.logger.info('writing the word-ID table...')
        self.byte_buffer.write_int(wordid_table.tell(), 'int')
        self.byte_buffer.seek(0)
        io_out.write(self.byte_buffer.read())
        self.byte_buffer.clear()

        wordid_table.seek(0)
        io_out.write(wordid_table.read())
        self.__logging_size(wordid_table.tell() + 4)
        del wordid_table

        self.logger.info('writing the word parameters...')
        self.byte_buffer.write_int(len(self.entries), 'int')
        for entry in self.entries:
            self.byte_buffer.write_int(entry.parameters[0], 'short')
            self.byte_buffer.write_int(entry.parameters[1], 'short')
            self.byte_buffer.write_int(entry.parameters[2], 'short')
            self.byte_buffer.seek(0)
            io_out.write(self.byte_buffer.read())
            self.byte_buffer.clear()
        self.__logging_size(len(self.entries) * 6 + 4)
        self.write_wordinfo(io_out)

    def write_wordinfo(self, io_out):
        mark = io_out.tell()
        io_out.seek(mark * 4 + len(self.entries))
        offsets = JTypedByteBuffer()
        self.logger.info('writing the word_infos...')
        base = io_out.tell()
        for entry in self.entries:
            wi = entry.wordinfo
            offsets.write_int(io_out.tell(), 'int')
            self.write_string(wi.surface)
            self.write_stringlength(wi.length())
            self.byte_buffer.write_int(wi.pos_id, 'short')
            if wi.normalized_form == wi.surface:
                self.write_string('')
            else:
                self.write_string(wi.normalized_form)
            self.byte_buffer.write_int(wi.dictionary_form_word_id, 'int')
            if wi.reading_form == wi.surface:
                self.write_string('')
            else:
                self.write_string(wi.reading_form)

            self.write_intarray(self.parse_splitinfo(entry.aunit_split_string))
            self.write_intarray(self.parse_splitinfo(entry.bunit_split_string))
            self.write_intarray(self.parse_splitinfo(entry.cunit_split_string))
            self.write_intarray(wi.synonym_group_ids)
            self.byte_buffer.seek(0)
            io_out.write(self.byte_buffer.read())
            self.byte_buffer.clear()
        self.__logging_size(io_out.tell() - base)
        self.logger.info('writing word_info offsets...')
        io_out.seek(mark)
        offsets.seek(0)
        io_out.write(offsets.read())
        self.__logging_size(offsets.tell())

    def decode(self, str_):
        def replace(match):
            uni_text = match.group()
            uni_text = uni_text.replace('{', '').replace('}', '')
            if len(uni_text) > 6:
                uni_text = ('\\U000{}'.format(uni_text[2:]))
            return uni_text.encode('ascii').decode('unicode-escape')
        return re.sub(self.__PATTERN_UNICODE_LITERAL, replace, str_)

    def check_splitinfo_format(self, str_):
        if str_.count('/') + 1 > self.__ARRAY_MAX_LENGTH:
            raise ValueError('too many units')

    def parse_splitinfo(self, info):
        if info == '*':
            return []
        words = info.split('/')
        if len(words) > self.__ARRAY_MAX_LENGTH:
            raise ValueError('too many units')
        ids = []
        for word in words:
            if self.__is_id(word):
                ids.append(self.parse_id(word))
            else:
                ids.append(self.word_to_id(word))
                if ids[-1] < 0:
                    return ValueError('not found such a word')
        return ids

    @staticmethod
    def __is_id(text):
        return re.match(r'U?\d+', text)

    def parse_id(self, text):
        if text.startswith('U'):
            id_ = int(text[1:])
            if self.is_user_dictionary:
                id_ |= (1 << 28)
        else:
            id_ = int(text)
        self.check_wordid(id_)
        return id_

    def word_to_id(self, text):
        cols = text.split(',')
        if len(cols) < 8:
            raise ValueError('too few columns')
        headword = self.decode(cols[0])
        pos_id = self.get_posid([cols[i] for i in range(1, 7)])
        if pos_id < 0:
            raise ValueError('invalid part of speech')
        reading = self.decode(cols[7])
        return self.get_wordid(headword, pos_id, reading)

    def get_wordid(self, headword, pos_id, reading_form):
        for i in range(len(self.entries)):
            info = self.entries[i].wordinfo
            if info.surface == headword \
                    and info.pos_id == pos_id \
                    and info.reading_form == reading_form:
                return i
        return -1

    def check_wordid(self, wid):
        if wid < 0 or wid >= len(self.entries):
            raise ValueError('invalid word ID')

    def parse_synonym_group_ids(self, text):
        if text == '*':
            return []
        synonym_group_ids = text.split('/')
        if len(synonym_group_ids) > self.__ARRAY_MAX_LENGTH:
            raise ValueError("too many units")
        return [int(synonym_group_id) for synonym_group_id in synonym_group_ids]

    def write_string(self, text):
        len_ = 0
        for c in text:
            if 0x10000 <= ord(c) <= 0x10FFFF:
                len_ += 2
            else:
                len_ += 1
        self.write_stringlength(len_)
        self.byte_buffer.write_str(text)

    def write_stringlength(self, len_):
        if len_ <= self.__BYTE_MAX_VALUE:
            self.byte_buffer.write_int(len_, 'byte')
        else:
            self.byte_buffer.write_int((len_ >> 8) | 0x80, 'byte')
            self.byte_buffer.write_int((len_ & 0xFF), 'byte')

    def write_intarray(self, array):
        self.byte_buffer.write_int(len(array), 'byte')
        for item in array:
            self.byte_buffer.write_int(item, 'int')

    def __logging_size(self, size):
        self.logger.info('{} bytes\n'.format(size))



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sudachipy import config
from sudachipy.dictionarylib.grammar import Grammar

from .editconnectioncost import EditConnectionCostPlugin


class InhibitConnectionPlugin(EditConnectionCostPlugin):
    """ A Edit Connection Cost Plugin for inhibiting the connections.

    The following is an example of settings.

    ``
    {
        {
            "class" : "sudachipy.plugin.connect_cost.InhibitConnectionPlugin",
            "inhibitedPair" : [ [ 0, 233 ], [435, 332] ]
        }
    }
    ``

    Attributes:
        _inhibit_pairs: a list of int pairs. At each pair, the first one is right-ID
        of the left node and the second one is left-ID of the right node in a connection.

    """

    def __init__(self):
        self._inhibit_pairs = []

    def set_up(self, grammar: Grammar) -> None:
        if 'inhibitedPair' in config.settings:
            self._inhibit_pairs = config.settings['inhibitedPair']

    def edit(self, grammar: Grammar) -> None:
        for pair in self._inhibit_pairs:
            if len(pair) < 2:
                continue
            self.inhibit_connection(grammar, pair[0], pair[1])



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
import unittest

from sudachipy.dictionarylib.categorytype import CategoryType
from sudachipy.plugin.oov import MeCabOovPlugin

from tests import mock_grammar, mock_inputtext


class TestMecabOOVPlugin(unittest.TestCase):

    def setUp(self):
        self.plugin = MeCabOovPlugin()
        oov1 = MeCabOovPlugin.OOV()
        oov1.pos_id = 1
        oov2 = MeCabOovPlugin.OOV()
        oov2.pos_id = 2
        self.plugin.oov_list[CategoryType.KANJI] = [oov1]
        self.plugin.oov_list[CategoryType.KANJINUMERIC] = [oov1, oov2]
        self.mocked_input_text = mock_inputtext.mocked_input_text
        mock_inputtext.set_text('あいうえお')

        self.test_dir = tempfile.mkdtemp()

    def test_provide_oov000(self):
        cinfo = MeCabOovPlugin.CategoryInfo()
        cinfo.type_ = CategoryType.KANJI
        cinfo.is_invoke = False
        cinfo.is_group = False
        cinfo.length = 0
        self.plugin.categories[CategoryType.KANJI] = cinfo

        mock_inputtext.set_category_type(0, 3, CategoryType.KANJI)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, False)
        self.assertEqual(0, len(nodes))

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, True)
        self.assertEqual(0, len(nodes))

    def test_provide_oov100(self):
        cinfo = MeCabOovPlugin.CategoryInfo()
        cinfo.type_ = CategoryType.KANJI
        cinfo.is_invoke = True
        cinfo.is_group = False
        cinfo.length = 0
        self.plugin.categories[CategoryType.KANJI] = cinfo

        mock_inputtext.set_category_type(0, 3, CategoryType.KANJI)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, False)
        self.assertEqual(0, len(nodes))

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, True)
        self.assertEqual(0, len(nodes))

    def test_provide_oov010(self):
        cinfo = MeCabOovPlugin.CategoryInfo()
        cinfo.type_ = CategoryType.KANJI
        cinfo.is_invoke = False
        cinfo.is_group = True
        cinfo.length = 0
        self.plugin.categories[CategoryType.KANJI] = cinfo

        mock_inputtext.set_category_type(0, 3, CategoryType.KANJI)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, False)
        self.assertEqual(1, len(nodes))

        n = nodes[0]
        self.assertEqual('あいう', n.get_word_info().surface)
        self.assertEqual(3, n.get_word_info().length())
        self.assertEqual(1, n.get_word_info().pos_id)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, True)
        self.assertEqual(0, len(nodes))

    def test_provide_oov110(self):
        cinfo = MeCabOovPlugin.CategoryInfo()
        cinfo.type_ = CategoryType.KANJI
        cinfo.is_invoke = True
        cinfo.is_group = True
        cinfo.length = 0
        self.plugin.categories[CategoryType.KANJI] = cinfo

        mock_inputtext.set_category_type(0, 3, CategoryType.KANJI)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, False)
        self.assertEqual(1, len(nodes))

        n = nodes[0]
        self.assertEqual('あいう', n.get_word_info().surface)
        self.assertEqual(3, n.get_word_info().length())
        self.assertEqual(1, n.get_word_info().pos_id)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, True)
        self.assertEqual(1, len(nodes))

    def test_provide_oov002(self):
        cinfo = MeCabOovPlugin.CategoryInfo()
        cinfo.type_ = CategoryType.KANJI
        cinfo.is_invoke = False
        cinfo.is_group = False
        cinfo.length = 2
        self.plugin.categories[CategoryType.KANJI] = cinfo

        mock_inputtext.set_category_type(0, 3, CategoryType.KANJI)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, False)
        self.assertEqual(2, len(nodes))

        n = nodes[0]
        self.assertEqual('あ', n.get_word_info().surface)
        self.assertEqual(1, n.get_word_info().length())
        self.assertEqual(1, n.get_word_info().pos_id)

        n = nodes[1]
        self.assertEqual('あい', n.get_word_info().surface)
        self.assertEqual(2, n.get_word_info().length())
        self.assertEqual(1, n.get_word_info().pos_id)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, True)
        self.assertEqual(0, len(nodes))

    def test_provide_oov012(self):
        cinfo = MeCabOovPlugin.CategoryInfo()
        cinfo.type_ = CategoryType.KANJI
        cinfo.is_invoke = False
        cinfo.is_group = True
        cinfo.length = 2
        self.plugin.categories[CategoryType.KANJI] = cinfo

        mock_inputtext.set_category_type(0, 3, CategoryType.KANJI)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, False)
        self.assertEqual(3, len(nodes))

        n = nodes[0]
        self.assertEqual('あいう', n.get_word_info().surface)
        self.assertEqual(3, n.get_word_info().length())
        self.assertEqual(1, n.get_word_info().pos_id)

        n = nodes[1]
        self.assertEqual('あ', n.get_word_info().surface)
        self.assertEqual(1, n.get_word_info().length())
        self.assertEqual(1, n.get_word_info().pos_id)

        n = nodes[2]
        self.assertEqual('あい', n.get_word_info().surface)
        self.assertEqual(2, n.get_word_info().length())
        self.assertEqual(1, n.get_word_info().pos_id)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, True)
        self.assertEqual(0, len(nodes))

    def test_provide_oov112(self):
        cinfo = MeCabOovPlugin.CategoryInfo()
        cinfo.type_ = CategoryType.KANJI
        cinfo.is_invoke = True
        cinfo.is_group = True
        cinfo.length = 2
        self.plugin.categories[CategoryType.KANJI] = cinfo

        mock_inputtext.set_category_type(0, 3, CategoryType.KANJI)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, False)
        self.assertEqual(3, len(nodes))

        n = nodes[0]
        self.assertEqual('あいう', n.get_word_info().surface)
        self.assertEqual(3, n.get_word_info().length())
        self.assertEqual(1, n.get_word_info().pos_id)

        n = nodes[1]
        self.assertEqual('あ', n.get_word_info().surface)
        self.assertEqual(1, n.get_word_info().length())
        self.assertEqual(1, n.get_word_info().pos_id)

        n = nodes[2]
        self.assertEqual('あい', n.get_word_info().surface)
        self.assertEqual(2, n.get_word_info().length())
        self.assertEqual(1, n.get_word_info().pos_id)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, True)
        self.assertEqual(3, len(nodes))

    def test_provide_oov006(self):
        cinfo = MeCabOovPlugin.CategoryInfo()
        cinfo.type_ = CategoryType.KANJI
        cinfo.is_invoke = False
        cinfo.is_group = False
        cinfo.length = 6
        self.plugin.categories[CategoryType.KANJI] = cinfo

        mock_inputtext.set_category_type(0, 3, CategoryType.KANJI)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, False)
        self.assertEqual(3, len(nodes))

        n = nodes[0]
        self.assertEqual('あ', n.get_word_info().surface)
        self.assertEqual(1, n.get_word_info().length())
        self.assertEqual(1, n.get_word_info().pos_id)

        n = nodes[1]
        self.assertEqual('あい', n.get_word_info().surface)
        self.assertEqual(2, n.get_word_info().length())
        self.assertEqual(1, n.get_word_info().pos_id)

        n = nodes[2]
        self.assertEqual('あいう', n.get_word_info().surface)
        self.assertEqual(3, n.get_word_info().length())
        self.assertEqual(1, n.get_word_info().pos_id)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, True)
        self.assertEqual(0, len(nodes))

    def test_provide_oov_multi_oov(self):
        cinfo = MeCabOovPlugin.CategoryInfo()
        cinfo.type_ = CategoryType.KANJINUMERIC
        cinfo.is_invoke = False
        cinfo.is_group = True
        cinfo.length = 0
        self.plugin.categories[CategoryType.KANJINUMERIC] = cinfo

        mock_inputtext.set_category_type(0, 3, CategoryType.KANJINUMERIC)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, False)
        self.assertEqual(2, len(nodes))

        n = nodes[0]
        self.assertEqual('あいう', n.get_word_info().surface)
        self.assertEqual(3, n.get_word_info().length())
        self.assertEqual(1, n.get_word_info().pos_id)

        n = nodes[1]
        self.assertEqual('あいう', n.get_word_info().surface)
        self.assertEqual(3, n.get_word_info().length())
        self.assertEqual(2, n.get_word_info().pos_id)

    def test_provide_oov_without_cinfo(self):
        mock_inputtext.set_category_type(0, 3, CategoryType.KANJI)
        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, False)
        self.assertEqual(0, len(nodes))

    def test_provide_oov_without_oov_list(self):
        cinfo = MeCabOovPlugin.CategoryInfo()
        cinfo.type_ = CategoryType.HIRAGANA
        cinfo.is_invoke = False
        cinfo.is_group = True
        cinfo.length = 0
        self.plugin.categories[CategoryType.HIRAGANA] = cinfo

        mock_inputtext.set_category_type(0, 3, CategoryType.HIRAGANA)

        nodes = self.plugin.provide_oov(self.mocked_input_text, 0, False)
        self.assertEqual(0, len(nodes))

    def test_read_character_property(self):
        input_ = os.path.join(self.test_dir, 'test.txt')
        with open(input_, 'w') as wf:
            wf.write("#\n  \nDEFAULT 0 1 2\nALPHA 1 0 0\n0x0000...0x0002 ALPHA")
        plugin = MeCabOovPlugin()
        plugin.read_character_property(input_)
        self.assertFalse(plugin.categories[CategoryType.DEFAULT].is_invoke)
        self.assertTrue(plugin.categories[CategoryType.DEFAULT].is_group)
        self.assertEqual(2, plugin.categories[CategoryType.DEFAULT].length)

    def test_read_character_property_with_too_few_columns(self):
        input_ = os.path.join(self.test_dir, 'test.txt')
        with open(input_, 'w') as wf:
            wf.write("DEFAULT 0 1\n")
        plugin = MeCabOovPlugin()
        with self.assertRaises(ValueError) as cm:
            plugin.read_character_property(input_)
        self.assertEqual('invalid format at line 1', cm.exception.args[0])

    def test_read_character_property_with_undefined_type(self):
        input_ = os.path.join(self.test_dir, 'test.txt')
        with open(input_, 'w') as wf:
            wf.write("FOO 0 1 2\n")
        plugin = MeCabOovPlugin()
        with self.assertRaises(ValueError) as cm:
            plugin.read_character_property(input_)
        self.assertEqual('`FOO` is invalid type at line 1', cm.exception.args[0])

    def test_read_character_property_duplicate_definitions(self):
        input_ = os.path.join(self.test_dir, 'test.txt')
        with open(input_, 'w') as wf:
            wf.write("DEFAULT 0 1 2\nDEFAULT 1 1 2")
        plugin = MeCabOovPlugin()
        with self.assertRaises(ValueError) as cm:
            plugin.read_character_property(input_)
        self.assertEqual('`DEFAULT` is already defined at line 2', cm.exception.args[0])

    def test_read_oov(self):
        oov = os.path.join(self.test_dir, 'test.txt')
        with open(oov, 'w') as wf:
            wf.write("DEFAULT,1,2,3,補助記号,一般,*,*,*,*\n")
            wf.write("DEFAULT,3,4,5,補助記号,一般,*,*,*,*\n")
        plugin = MeCabOovPlugin()
        plugin.categories[CategoryType.DEFAULT] = MeCabOovPlugin.CategoryInfo()
        plugin.read_oov(oov, mock_grammar.mocked_grammar)
        self.assertEqual(1, len(plugin.oov_list))
        self.assertEqual(2, len(plugin.oov_list[CategoryType.DEFAULT]))
        self.assertEqual(1, plugin.oov_list[CategoryType.DEFAULT][0].left_id)
        self.assertEqual(2, plugin.oov_list[CategoryType.DEFAULT][0].right_id)
        self.assertEqual(3, plugin.oov_list[CategoryType.DEFAULT][0].cost)
        self.assertEqual(0, plugin.oov_list[CategoryType.DEFAULT][0].pos_id)

    def test_read_oov_with_too_few_columns(self):
        input_ = os.path.join(self.test_dir, 'test.txt')
        with open(input_, 'w') as wf:
            wf.write("DEFAULT,1,2,3\n")
        plugin = MeCabOovPlugin()
        plugin.categories[CategoryType.DEFAULT] = MeCabOovPlugin.CategoryInfo()
        with self.assertRaises(ValueError) as cm:
            plugin.read_oov(input_, mock_grammar.mocked_grammar)
        self.assertEqual('invalid format at line 1', cm.exception.args[0])

    def test_read_oov_with_undefined_type(self):
        input_ = os.path.join(self.test_dir, 'test.txt')
        with open(input_, 'w') as wf:
            wf.write("FOO,1,2,3,補助記号,一般,*,*,*,*\n")
        plugin = MeCabOovPlugin()
        plugin.categories[CategoryType.DEFAULT] = MeCabOovPlugin.CategoryInfo()
        with self.assertRaises(ValueError) as cm:
            plugin.read_oov(input_, mock_grammar.mocked_grammar)
        self.assertEqual('`FOO` is invalid type at line 1', cm.exception.args[0])

    def test_read_oov_with_category_not_in_character_property(self):
        input_ = os.path.join(self.test_dir, 'test.txt')
        with open(input_, 'w') as wf:
            wf.write("ALPHA,1,2,3,補助記号,一般,*,*,*,*\n")
        plugin = MeCabOovPlugin()
        plugin.categories[CategoryType.DEFAULT] = MeCabOovPlugin.CategoryInfo()
        with self.assertRaises(ValueError) as cm:
            plugin.read_oov(input_, mock_grammar.mocked_grammar)
        self.assertEqual('`ALPHA` is undefined at line 1', cm.exception.args[0])


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#from sudachipy.plugin.oov import simple_oov_plugin
#from sudachipy.plugin.oov import mecab_oov_plugin
#from sudachipy.plugin.path_rewrite import join_numeric_plugin
#from sudachipy.plugin.path_rewrite import join_katakana_oov_plugin
#from sudachipy.plugin.input_text import default_input_text_plugin
#from .connect_cost.inhibitconnectioncost import InhibitConnectionPlugin
#from . import oov
#from . import input_text
#from . import path_rewrite



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

from .dictionarylib.grammar import Grammar
from .latticenode cimport LatticeNode

cdef class Lattice:

    def __init__(self, grammar: Grammar):
        self.size = 0
        self.capacity = 0


        self.end_lists = []
        self.grammar = grammar
        self.eos_params = grammar.get_eos_parameter()
        cdef LatticeNode bos_node = LatticeNode()
        bos_params = grammar.get_bos_parameter()
        bos_node.set_parameter(bos_params[0], bos_params[1], bos_params[2])
        bos_node.is_connected_to_bos = True
        self.end_lists.append([bos_node])
        self.connect_costs = self.grammar._matrix_view

    cpdef void resize(self, int size):
        if size > self.capacity:
            self.expand(size)
        self.size = size
        self.eos_node = LatticeNode()
        self.eos_node.set_parameter(self.eos_params[0], self.eos_params[1], self.eos_params[2])
        self.eos_node.begin = self.eos_node.end = size

    def clear(self) -> None:
        for i in range(1, self.size + 1):
            self.end_lists[i].clear()
        self.size = 0
        self.eos_node = None

    def expand(self, new_size: int) -> None:
        expand_list = [[] for _ in range(self.size, new_size)]
        self.end_lists.extend(expand_list)
        self.capacity = new_size

    def get_nodes_with_end(self, end: int) -> List[LatticeNode]:
        return self.end_lists[end]

    def get_nodes(self, begin: int, end: int) -> List[LatticeNode]:
        return [node for node in self.end_lists[end] if node.begin == begin]

    def get_minumum_node(self, begin: int, end: int) -> Optional[LatticeNode]:
        nodes = self.get_nodes(begin, end)
        if not nodes:
            return None
        min_arg = nodes[0]
        for node in nodes[1:]:
            if node.cost < min_arg.cost:
                min_arg = node
        return min_arg

    cpdef void insert(self, int begin, int end, LatticeNode node):
        self.end_lists[end].append(node)
        node.begin = begin
        node.end = end
        self.connect_node(node)

    def remove(self, begin: int, end: int, node: LatticeNode) -> None:
        self.end_lists[end].remove(node)

    @staticmethod
    def create_node() -> LatticeNode:
        return LatticeNode()

    def has_previous_node(self, index: int) -> bool:
        return bool(self.end_lists[index])

    cdef void connect_node(self, LatticeNode r_node):
        begin = r_node.begin
        r_node.total_cost = INT_MAX

        cdef LatticeNode l_node
        cdef int connect_cost
        for l_node in self.end_lists[begin]:
            if not l_node.is_connected_to_bos:
                continue
            connect_cost = self.connect_costs[r_node.left_id, l_node.right_id]

            # 0x7fff == Grammar.INHIBITED_CONNECTION:
            if connect_cost == 0x7fff:
                continue
            cost = l_node.total_cost + connect_cost
            if cost < r_node.total_cost:
                r_node.total_cost = cost
                r_node.best_previous_node = l_node

        r_node.is_connected_to_bos = r_node.best_previous_node is not None
        r_node.total_cost += r_node.cost

    cdef void connect_eos_node(self):
        self.connect_node(self.eos_node)

    def get_best_path(self) -> List[LatticeNode]:
        # self.connect_node(self.eos_node)
        if not self.eos_node.is_connected_to_bos:    # EOS node
            raise AttributeError("EOS is not connected to BOS")
        result = []
        node = self.eos_node.best_previous_node
        while node is not self.end_lists[0][0]:
            result.append(node)
            node = node.best_previous_node
        return list(reversed(result))

    def dump(self, logger):
        if logger.disabled:
            return
        index = 0
        for i in range(self.size + 1, -1, -1):
            r_nodes = self.end_lists[i] if i <= self.size else [self.eos_node]
            for r_node in r_nodes:
                surface = '(null)'
                pos = 'BOS/EOS'

                if r_node.is_defined():
                    wi = r_node.get_word_info()
                    surface = wi.surface
                    pos_id = wi.pos_id
                    pos = '(null)'
                    if pos_id >= 0:
                        pos = ','.join(self.grammar.get_part_of_speech_string(pos_id))

                costs = []
                for l_node in self.end_lists[r_node.get_begin()]:
                    cost = self.grammar.get_connect_cost(l_node.get_right_id(), r_node.get_left_id())
                    costs.append(str(cost))
                index += 1

                logger.info('%d: %d %d %s(%d) %s %d %d %d: %s' %
                            (index, r_node.get_begin(), r_node.get_end(),
                             surface, r_node.get_word_id(), pos, r_node.get_left_id(),
                             r_node.get_right_id(), r_node.get_path_cost(), ' '.join(costs)))



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Iterator, List

from .wordinfo import WordInfo


class Lexicon(ABC):

    Itr = Iterator[List[int]]

    @abstractmethod
    def lookup(self, text: str, offset: int) -> Itr:  # noqa: F821
        raise NotImplementedError

    @abstractmethod
    def get_word_id(self, headword: str, pos_id: int, reading_form: str) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_left_id(self, word_id: int) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_right_id(self, word_id: int) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_cost(self, word_id: int) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_word_info(self, word_id: int) -> 'WordInfo':
        raise NotImplementedError

    @abstractmethod
    def get_dictionary_id(self, word_id: int) -> int:
        raise NotImplementedError

    @abstractmethod
    def size(self) -> int:
        raise NotImplementedError



# cython: profile=True

# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .dictionarylib.wordinfo import WordInfo

__NULL_SURFACE = '(null)'
UNK =\
    WordInfo(__NULL_SURFACE, 0, -1, __NULL_SURFACE, -1,
             __NULL_SURFACE, __NULL_SURFACE, [], [], [], [])

cdef class LatticeNode:

    def __init__(self, lexicon=None, left_id=None, right_id=None, cost=None, word_id=None):

        self.begin = 0
        self.end = 0
        self.word_id = 0
        self._is_oov = False
        self.best_previous_node = None
        self.is_connected_to_bos = False
        self.extra_word_info = None

        self._is_defined = True
        if lexicon is left_id is right_id is cost is word_id is None:
            self._is_defined = False
            return
        self.lexicon = lexicon
        self.left_id = left_id
        self.right_id = right_id
        self.cost = cost
        self.word_id = word_id

    def set_parameter(self, left_id: int, right_id: int, cost: int) -> None:
        self.left_id = left_id
        self.right_id = right_id
        self.cost = cost

    def get_begin(self) -> int:
        return self.begin

    def set_begin(self, begin) -> None:
        self.begin = begin

    def get_end(self) -> int:
        return self.end

    def set_end(self, end) -> None:
        self.end = end

    def set_range(self, begin: int, end: int) -> None:
        self.begin = begin
        self.end = end

    def is_oov(self):
        return self._is_oov

    def set_oov(self):
        self._is_oov = True
    
    def is_defined(self):
        return self._is_defined
    
    def set_defined(self):
        self._is_defined = True
    
    def get_word_info(self) -> WordInfo:
        if not self._is_defined:
            return UNK
        if self.extra_word_info:
            return self.extra_word_info
        return self.lexicon.get_word_info(self.word_id)

    def set_word_info(self, word_info: WordInfo) -> None:
        self.extra_word_info = word_info
        self._is_defined = True

    def get_path_cost(self) -> int:
        return self.cost
    
    def get_left_id(self) -> int:
        return self.left_id
    
    def get_right_id(self) -> int:
        return self.right_id

    def get_word_id(self) -> int:
        return self.word_id

    def get_dictionary_id(self) -> int:
        if not self._is_defined or self.extra_word_info:
            return -1
        return self.lexicon.get_dictionary_id(self.word_id)  # self.word_id >> 28

    def __str__(self):
        surface = "(None)"
        if self.word_id >= 0 or self.extra_word_info:
            surface = self.get_word_info().surface

        return "{} {} {}({}) {} {} {}".format(
            self.get_begin(), self.get_end(), surface, self.word_id, self.left_id, self.right_id, self.cost
        )



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import latticenode
from . import morpheme
from . import tokenizer


class MorphemeList:

    @classmethod
    def empty(cls):
        return MorphemeList(None, None, None, [])

    def __init__(self, input_, grammar, lexicon, path):
        self.input_text = input_
        self.grammar = grammar
        self.lexicon = lexicon
        self.path = path

    def __getitem__(self, index):
        n_morphs = len(self.path)
        if index >= n_morphs or index < -n_morphs:
            raise IndexError("Morpheme list index out of range")
        if index < 0:
            return morpheme.Morpheme(self, n_morphs + index)
        return morpheme.Morpheme(self, index)

    def __len__(self):
        return len(self.path)

    def __iter__(self):
        for index in range(len(self.path)):
            yield morpheme.Morpheme(self, index)
        return

    def __str__(self):
        return ''.join([mm.surface() for mm in self])

    def get_begin(self, index):
        return self.input_text.get_original_index(self.path[index].get_begin())

    def get_end(self, index):
        return self.input_text.get_original_index(self.path[index].get_end())

    def get_surface(self, index):
        begin = self.get_begin(index)
        end = self.get_end(index)
        return self.input_text.get_original_text()[begin:end]

    def get_word_info(self, index):
        return self.path[index].get_word_info()

    def split(self, mode, index, wi):
        if mode is tokenizer.Tokenizer.SplitMode.A:
            word_ids = wi.a_unit_split
        elif mode is tokenizer.Tokenizer.SplitMode.B:
            word_ids = wi.b_unit_split
        else:
            return [self.__getitem__(index)]

        if len(word_ids) == 0 or len(word_ids) == 1:
            return [self.__getitem__(index)]

        offset = self.path[index].get_begin()
        nodes = []
        for wid in word_ids:
            n = latticenode.LatticeNode(self.lexicon, 0, 0, 0, wid)
            n.set_begin(offset)
            offset += n.get_word_info().head_word_length
            n.set_end(offset)
            nodes.append(n)

        return MorphemeList(self.input_text, self.grammar, self.lexicon, nodes)

    def is_oov(self, index):
        return self.path[index].is_oov()

    def get_internal_cost(self):
        return self.path[-1].get_path_cost() - self.path[0].get_path_cost()

    def size(self):
        return len(self.path)



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import InputTextPlugin


class ProlongedSoundMarkInputTextPlugin(InputTextPlugin):

    def __init__(self, json_obj):
        self._psm_set = set()
        self._replace_symbol = 'ー'
        if not json_obj:
            return
        if 'prolongedSoundMarks' in json_obj:
            self._psm_set = set([ord(psm) for psm in json_obj['prolongedSoundMarks']])
        if 'replacementSymbol' in json_obj:
            self._replace_symbol = json_obj['replacementSymbol']

    def set_up(self) -> None:
        pass

    def rewrite(self, builder: InputTextPlugin.Builder) -> None:
        text = builder.get_text()
        n = len(text)
        offset = 0
        is_psm = False
        m_start_idx = n
        for i in range(n):
            cp = ord(text[i])
            if not is_psm and cp in self._psm_set:
                is_psm = True
                m_start_idx = i
            elif is_psm and cp not in self._psm_set:
                if i - m_start_idx > 1:
                    builder.replace(m_start_idx - offset, i - offset, self._replace_symbol)
                    offset += i - m_start_idx - 1
                is_psm = False
        if is_psm and n - m_start_idx > 1:
            builder.replace(m_start_idx - offset, n - offset, self._replace_symbol)



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from io import StringIO


class NumericParser(object):

    class StringNumber(object):

        def __init__(self):
            self._significand = StringIO()
            self._scale = 0
            self._point = -1
            self._is_all_zero = True

        def clear(self) -> None:
            self._significand.seek(0)
            self._significand.truncate(0)
            self._scale = 0
            self._point = -1
            self._is_all_zero = True

        def append(self, i: int) -> None:
            if i != 0:
                self._is_all_zero = False
            self._significand.write(str(i))

        def shift_scale(self, i: int) -> None:
            if self.is_zero():
                self._significand.write('1')
            self._scale += i

        def add(self, number: 'StringNumber') -> bool:  # noqa: F821
            if number.is_zero():
                return True

            if self.is_zero():
                self._significand.write(number.significand_value())
                self._scale = number.scale
                self._point = number.point
                return True

            self.normalize_scale()
            length = number.int_length()
            if self._scale >= length:
                self.fill_zero(self._scale - length)
                if number.point >= 0:
                    self._point = self._significand.tell() + number.point
                self._significand.write(number.significand_value())
                self._scale = number.scale
                return True

            return False

        def set_point(self) -> bool:
            if self._scale == 0 and self._point < 0:
                self._point = self._significand.tell()
                return True
            return False

        def int_length(self) -> int:
            self.normalize_scale()
            if self._point >= 0:
                return self._point
            return self._significand.tell() + self._scale

        def is_zero(self):
            return self._significand.tell() == 0

        def normalize_scale(self) -> None:
            if self._point < 0:
                return
            n_scale = self._significand.tell() - self._point
            if n_scale > self._scale:
                self._point += self._scale
                self._scale = 0
            else:
                self._scale -= n_scale
                self._point = -1

        def fill_zero(self, length: int) -> None:
            if length > 0:
                self._significand.write('0' * length)

        @property
        def scale(self) -> int:
            return self._scale

        @property
        def point(self) -> int:
            return self._point

        @property
        def is_all_zero(self) -> bool:
            return self._is_all_zero

        def significand_value(self) -> str:
            return self._significand.getvalue()

        def insert_significand(self, idx: int, text: str) -> None:
            self._significand.seek(idx)
            post_fix = self._significand.read()
            self._significand.seek(idx)
            self._significand.write(text)
            self._significand.write(post_fix)

        def __str__(self):
            if self.is_zero():
                return '0'

            self.normalize_scale()
            if self._scale > 0:
                self.fill_zero(self._scale)
            elif self._point >= 0:
                self.insert_significand(self._point, '.')
                if self._point == 0:
                    self.insert_significand(0, '0')
                i = self._significand.tell() - 1
                self._significand.seek(i)
                while i >= 0 and self._significand.read(1) == '0':
                    i -= 1
                    self._significand.seek(i)
                self._significand.truncate(i + 1)
                self._significand.seek(i)
                if self._significand.read(1) == '.':
                    self._significand.truncate(i)

            return self._significand.getvalue()

    class Error(Enum):
        NONE = 1
        POINT = 2
        COMMA = 3
        OTHER = 4

    _digit_length = 0
    _is_first_digit = True
    _has_comma = False
    _has_hanging_point = False
    _error_state = Error.NONE
    _total = None
    _sub_total = None
    _tmp = None

    def __init__(self):
        self._char_to_num = {}
        for i in range(10):
            self._char_to_num[str(i)] = i
        self._char_to_num['〇'] = 0
        self._char_to_num['一'] = 1
        self._char_to_num['二'] = 2
        self._char_to_num['三'] = 3
        self._char_to_num['四'] = 4
        self._char_to_num['五'] = 5
        self._char_to_num['六'] = 6
        self._char_to_num['七'] = 7
        self._char_to_num['八'] = 8
        self._char_to_num['九'] = 9
        self._char_to_num['十'] = -1
        self._char_to_num['百'] = -2
        self._char_to_num['千'] = -3
        self._char_to_num['万'] = -4
        self._char_to_num['億'] = -8
        self._char_to_num['兆'] = -12

        self._total = self.StringNumber()
        self._sub_total = self.StringNumber()
        self._tmp = self.StringNumber()
        self.clear()

    def clear(self) -> None:
        self._digit_length = 0
        self._is_first_digit = True
        self._has_comma = False
        self._has_hanging_point = False
        self._error_state = self.Error.NONE
        self._total.clear()
        self._sub_total.clear()
        self._tmp.clear()

    def append(self, char: str) -> bool:
        if len(char) > 1:
            raise ValueError('char must be single character string')
        if char == '.':
            self._has_hanging_point = True
            if self._is_first_digit:
                self._error_state = self.Error.POINT
                return False
            if self._has_comma and not self.check_comma():
                self._error_state = self.Error.COMMA
                return False
            if not self._tmp.set_point():
                self._error_state = self.Error.POINT
                return False
            self._has_comma = False
            return True
        if char == ',':
            if not self.check_comma():
                self._error_state = self.Error.COMMA
                return False
            self._has_comma = True
            self._digit_length = 0
            return True

        if char not in self._char_to_num:
            return False

        num = self._char_to_num[char]
        if self.is_small_unit(num):
            self._tmp.shift_scale(-num)
            if not self._sub_total.add(self._tmp):
                return False
            self._tmp.clear()
            self._is_first_digit = True
            self._digit_length = 0
            self._has_comma = False
            return True
        if self.is_large_unit(num):
            if not self._sub_total.add(self._tmp) or self._sub_total.is_zero():
                return False
            self._sub_total.shift_scale(-num)
            if not self._total.add(self._sub_total):
                return False
            self._sub_total.clear()
            self._tmp.clear()
            self._is_first_digit = True
            self._digit_length = 0
            self._has_comma = False
            return True
        self._tmp.append(num)
        self._is_first_digit = False
        self._digit_length += 1
        self._has_hanging_point = False
        return True

    def done(self) -> bool:
        ret = self._sub_total.add(self._tmp) and self._total.add(self._sub_total)
        if self._has_hanging_point:
            self._error_state = self.Error.POINT
            return False
        if self._has_comma and self._digit_length != 3:
            self._error_state = self.Error.COMMA
            return False
        return ret

    def get_normalized(self) -> str:
        return str(self._total)

    def check_comma(self) -> bool:
        if self._is_first_digit:
            return False
        if not self._has_comma:
            return self._digit_length <= 3 and not self._tmp.is_zero() and not self._tmp.is_all_zero
        return self._digit_length == 3

    @staticmethod
    def is_small_unit(num: int):
        return -3 <= num < 0

    @staticmethod
    def is_large_unit(num: int):
        return num <= -4

    @property
    def error_state(self):
        return self._error_state



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class WordParameterList(object):

    ELEMENT_SIZE = 2 * 3
    ELEMENT_SIZE_AS_SHORT = 3

    def __init__(self, bytes_, offset):
        original_offset = bytes_.tell()
        bytes_.seek(offset)
        self.size = int.from_bytes(bytes_.read(4), 'little')
        array_offset = bytes_.tell()
        self._array_view = memoryview(bytes_)[array_offset: array_offset + self.size * self.ELEMENT_SIZE]
        self._array_view = self._array_view.cast('h')
        # self.is_copied = False
        bytes_.seek(original_offset)

    def __del__(self):
        self._array_view.release()

    def storage_size(self):
        return 4 + self.ELEMENT_SIZE * self.size

    def get_size(self):
        return self.size

    def get_left_id(self, word_id):
        return self._array_view[self.ELEMENT_SIZE_AS_SHORT * word_id]

    def get_right_id(self, word_id):
        return self._array_view[self.ELEMENT_SIZE_AS_SHORT * word_id + 1]

    def get_cost(self, word_id):
        return self._array_view[self.ELEMENT_SIZE_AS_SHORT * word_id + 2]

    def set_cost(self, word_id, cost):
        # bytes_ must be ACCESS_COPY mode
        self._array_view[self.ELEMENT_SIZE_AS_SHORT * word_id + 2] = cost



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

from sudachipy.config import settings
from sudachipy.dictionary import Dictionary
from sudachipy.plugin.path_rewrite import JoinKatakanaOovPlugin
from sudachipy.utf8inputtextbuilder import UTF8InputTextBuilder


class TestJoinKatakanaOOVPlugin(unittest.TestCase):

    def setUp(self):
        resource_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, 'resources')
        self.dict_ = Dictionary(os.path.join(resource_dir, 'sudachi.json'), resource_dir)
        self.tokenizer = self.dict_.create()
        self.plugin = JoinKatakanaOovPlugin(settings['pathRewritePlugin'][1])

    def test_katakana_length(self):
        # アイ, アイウ in the dictionary
        self.plugin._min_length = 0
        path = self.get_path('アイアイウ')
        self.assertEqual(2, len(path))

        self.plugin._min_length = 1
        path = self.get_path('アイアイウ')
        self.assertEqual(2, len(path))

        self.plugin._min_length = 2
        path = self.get_path('アイアイウ')
        self.assertEqual(2, len(path))

        self.plugin._min_length = 3
        path = self.get_path('アイアイウ')
        self.assertEqual(1, len(path))

    def test_pos(self):
        # アイアイウ is 名詞-固有名詞-地名-一般 in the dictionary
        self.plugin._min_length = 3
        path = self.get_path('アイアイウ')
        self.assertEqual(1, len(path))
        self.assertFalse(path[0].is_oov())

    def test_starts_with_middle(self):
        self.plugin._min_length = 3
        path = self.get_path('アイウアイアイウ')
        self.assertEqual(1, len(path))

    def test_starts_with_tail(self):
        self.plugin._min_length = 3
        path = self.get_path('アイウアイウアイ')
        self.assertEqual(1, len(path))

    def test_with_nooovbow(self):
        self.plugin._min_length = 3
        path = self.get_path('ァアイアイウ')
        self.assertEqual(2, len(path))
        self.assertEqual('ァ', path[0].get_word_info().surface)

        path = self.get_path('アイウァアイウ')
        self.assertEqual(1, len(path))

    def get_path(self, text: str):
        input_ = UTF8InputTextBuilder(text, self.tokenizer._grammar).build()
        self.tokenizer._build_lattice(input_)
        path = self.tokenizer._lattice.get_best_path()
        self.plugin.rewrite(input_, path, self.tokenizer._lattice)
        self.tokenizer._lattice.clear()
        return path


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import tempfile
import time
from io import StringIO
from logging import getLogger
from unittest import TestCase, mock

from sudachipy.dictionarylib import SYSTEM_DICT_VERSION_2
from sudachipy.dictionarylib.dictionarybuilder import DictionaryBuilder
from sudachipy.dictionarylib.dictionaryheader import DictionaryHeader
from sudachipy.dictionarylib.lexicon import Lexicon
from sudachipy.dictionarylib.userdictionarybuilder import UserDictionaryBuilder


class TestDictionaryBuilder(TestCase):

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.matrix_path = os.path.join(self.test_dir, 'matrix.txt')
        with open(self.matrix_path, 'w', encoding='utf-8') as wf:
            wf.write('1 1\n0 0 200\n')
        self.input_path = os.path.join(self.test_dir, 'input.txt')
        with open(self.input_path, 'w', encoding='utf-8') as wf:
            wf.write("東京都,0,0,0,東京都,名詞,固有名詞,地名,一般,*,*,ヒガシキョウト,東京都,*,B,\"東,名詞,普通名詞,一般,*,*,*,ヒガシ/2\",*,1/2,1/2\n")
            wf.write("東,-1,-1,0,東,名詞,普通名詞,一般,*,*,*,ヒガシ,ひがし,*,A,*,*,*,*\n")
            wf.write("京都,0,0,0,京都,名詞,固有名詞,地名,一般,*,*,キョウト,京都,*,A,*,*,*,*\n")
        self.logger = getLogger()
        self.logger.disabled = True

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_parseline(self):
        builder = DictionaryBuilder(logger=self.logger)
        entry = builder.parse_line(
            '京都,6,6,5293,京都,名詞,固有名詞,地名,一般,*,*,キョウト,京都,*,A,*,*,*,*'.split(','))
        self.assertEqual('京都', entry.headword)
        self.assertEqual([6, 6, 5293], entry.parameters)
        self.assertEqual(0, entry.wordinfo.pos_id)
        self.assertEqual('*', entry.aunit_split_string)
        self.assertEqual('*', entry.bunit_split_string)

        entry = builder.parse_line(
            '京都,-1,-1,0,京都,名詞,固有名詞,地名,一般,*,*,キョウト,京都,*,A,*,*,*,*'.split(','))
        self.assertIsNone(entry.headword)
        self.assertEqual(0, entry.wordinfo.pos_id)

        # Synonym GID field is optional
        entry = builder.parse_line(
            '京都,6,6,5293,京都,名詞,固有名詞,地名,一般,*,*,キョウト,京都,*,A,*,*,*'.split(','))
        self.assertEqual(0, len(entry.wordinfo.synonym_group_ids))

    def test_parse_line_invalid_columns(self):
        builder = DictionaryBuilder(logger=self.logger)
        with self.assertRaises(ValueError) as cm:
            builder.parse_line('京都,6,6,5293,京都,名詞,固有名詞,地名,一般,*,*,キョウト,京都,*,A,*,*'.split(','))
        self.assertEqual('invalid format', cm.exception.args[0])

    def test_parse_line_empty_headword(self):
        builder = DictionaryBuilder(logger=self.logger)
        with self.assertRaises(ValueError) as cm:
            builder.parse_line(',6,6,5293,京都,名詞,固有名詞,地名,一般,*,*,キョウト,京都,*,A,*,*,*,*'.split(','))
        self.assertEqual('headword is empty', cm.exception.args[0])

    def test_parse_line_toolong_headword(self):
        builder = DictionaryBuilder(logger=self.logger)
        x = 'a' * (32767 + 1)  # max value of short in Java + 1
        x = x + ',6,6,5293,京都,名詞,固有名詞,地名,一般,*,*,キョウト,京都,*,A,*,*,*,*'
        with self.assertRaises(ValueError) as cm:
            builder.parse_line(x.split(','))
        self.assertEqual('string is too long', cm.exception.args[0])

    def test_parse_line_toomany_split(self):
        builder = DictionaryBuilder(logger=self.logger)
        with self.assertRaises(ValueError) as cm:
            builder.parse_line('京都,6,6,5293,京都,名詞,固有名詞,地名,一般,*,*,キョウト,京都,*,B,0/1/2/3/4/5/6/7/8/9/0/0/1/2/3/4/5/6/7/8/9/0/0/1/2/3/4/5/6/7/8/9/0/0/1/2/3/4/5/6/7/8/9/0/0/1/2/3/4/5/6/7/8/9/0/0/1/2/3/4/5/6/7/8/9/0/0/1/2/3/4/5/6/7/8/9/0/0/1/2/3/4/5/6/7/8/9/0/0/1/2/3/4/5/6/7/8/9/0/0/1/2/3/4/5/6/7/8/9/0/0/1/2/3/4/5/6/7/8/9/0/0/1/2/3/4/5/6/7/8/9/0/0/1/2/3/4/5/6/7/8/9/0,*,*,*'.split(','))
        self.assertEqual('too many units', cm.exception.args[0])

    def test_parse_line_same_readingform(self):
        builder = DictionaryBuilder(logger=self.logger)
        entry = builder.parse_line('〒,6,6,5293,〒,名詞,普通名詞,一般,*,*,*,〒,〒,*,A,*,*,*,*'.split(','))
        self.assertEqual('〒', entry.wordinfo.reading_form)

    def test_add_to_trie(self):
        builder = DictionaryBuilder(logger=self.logger)
        builder.add_to_trie('abc', 0)
        builder.add_to_trie('abc', 1)
        builder.add_to_trie('abcd', 2)
        self.assertTrue(0 in builder.trie_keys['abc'.encode('utf-8')])
        self.assertTrue(1 in builder.trie_keys['abc'.encode('utf-8')])

    def test_convert_postable(self):
        builder = DictionaryBuilder(logger=self.logger)
        builder.convert_postable(['a,b,c,d,e,f', 'g,h,i,j,k,l'])
        self.assertEqual(2 + 3 * 12, builder.byte_buffer.tell())

    def test_convert_matrix(self):
        in_stream = StringIO('2 3\n0 0 0\n0 1 1\n0 2 2\n\n1 0 3\n1 1 4\n1 2 5\n')
        builder = DictionaryBuilder(logger=self.logger)
        matrix = builder.convert_matrix(in_stream)
        self.assertEqual(2, int.from_bytes(builder.byte_buffer.getvalue()[0:2], byteorder='little'))
        self.assertEqual(3, int.from_bytes(builder.byte_buffer.getvalue()[2:4], byteorder='little'))
        self.assertEqual(0, int.from_bytes(matrix.getvalue()[0:2], byteorder='little'))
        self.assertEqual(4, int.from_bytes(matrix.getvalue()[(2 + 1) * 2:(2 + 1) * 2 + 2], byteorder='little'))

    def test_decode(self):
        builder = DictionaryBuilder(logger=self.logger)
        self.assertEqual('a,c', builder.decode('a\\u002cc'))
        self.assertEqual('a,c', builder.decode('a\\u{002c}c'))
        self.assertEqual('a𠮟c', builder.decode('a\\u{20b9f}c'))

    def test_parse_splitinfo(self):
        builder = DictionaryBuilder(logger=self.logger)
        builder.entries.extend([None] * 4)
        self.assertEqual([], builder.parse_splitinfo('*'))
        self.assertEqual([1, 2, 3], builder.parse_splitinfo('1/2/3'))
        self.assertEqual(2, builder.parse_splitinfo('1/U2/3')[1])

        mocked_lexicon = mock.Mock(spec=Lexicon)
        mocked_lexicon.size.return_value = 4
        builder = UserDictionaryBuilder(None, mocked_lexicon)
        builder.entries += [None, None, None]
        self.assertEqual([1, 2 | 1 << 28, 3], builder.parse_splitinfo("1/U2/3"))

    def test_parse_splitinfo_invalid_wordid(self):
        builder = DictionaryBuilder(logger=self.logger)
        with self.assertRaises(ValueError) as cm:
            builder.parse_splitinfo('1/2/3')
        self.assertEqual('invalid word ID', cm.exception.args[0])

    def test_parse_splitinfo_invalid_wordid_userdict(self):
        mocked_lexicon = mock.Mock(spec=Lexicon)
        mocked_lexicon.size.return_value = 1
        builder = UserDictionaryBuilder(None, mocked_lexicon)
        with self.assertRaises(ValueError) as cm:
            builder.parse_splitinfo('0/U1')
        self.assertEqual('invalid word ID', cm.exception.args[0])

    def test_parse_splitinfo_invalid_system_wordid_in_userdict(self):
        mocked_lexicon = mock.Mock(spec=Lexicon)
        mocked_lexicon.size.return_value = 1
        builder = UserDictionaryBuilder(None, mocked_lexicon)
        builder.entries.append(None)
        with self.assertRaises(ValueError) as cm:
            builder.parse_splitinfo('1/U0')
        self.assertEqual('invalid word id', cm.exception.args[0])
        pass

    def test_write_string(self):
        builder = DictionaryBuilder(logger=self.logger)
        position = builder.byte_buffer.tell()
        builder.write_string('')
        self.assertEqual(0, builder.byte_buffer.getvalue()[position])
        self.assertEqual(position + 1, builder.byte_buffer.tell())

        position = builder.byte_buffer.tell()
        builder.write_string('あ𠮟')
        self.assertEqual(3, builder.byte_buffer.getvalue()[position])
        self.assertEqual('あ', builder.byte_buffer.getvalue()[position + 1: position + 3].decode('utf-16-le'))
        a = int.from_bytes(builder.byte_buffer.getvalue()[position + 3: position + 5], byteorder='little')
        b = int.from_bytes(builder.byte_buffer.getvalue()[position + 5: position + 7], byteorder='little')
        self.assertEqual(55362, a)  # \ud842
        self.assertEqual(57247, b)  # \udf94

        position = builder.byte_buffer.tell()
        long_str = '0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789'
        len_ = len(long_str)
        builder.write_string(long_str)
        self.assertEqual((len_ >> 8) | 0x80, builder.byte_buffer.getvalue()[position])
        self.assertEqual(len_ & 0xff, builder.byte_buffer.getvalue()[position + 1])
        self.assertEqual(position + 2 + 2 * len_, builder.byte_buffer.tell())

    def test_write_intarray(self):
        builder = DictionaryBuilder(logger=self.logger)
        position = builder.byte_buffer.tell()
        builder.write_intarray([])
        self.assertEqual(0, builder.byte_buffer.getvalue()[position])
        builder.write_intarray([1, 2, 3])
        self.assertEqual(3, builder.byte_buffer.getvalue()[position + 1])
        self.assertEqual(1, int.from_bytes(builder.byte_buffer.getvalue()[position + 2:position + 6], byteorder='little', signed=True))
        self.assertEqual(2, int.from_bytes(builder.byte_buffer.getvalue()[position + 6:position + 10], byteorder='little', signed=True))
        self.assertEqual(3, int.from_bytes(builder.byte_buffer.getvalue()[position + 10:position + 14], byteorder='little', signed=True))

    def test_build(self):
        out_path = os.path.join(self.test_dir, 'output.txt')

        out_stream = open(out_path, 'wb')
        lexicon_paths = [self.input_path]
        matrix_input_stream = open(self.matrix_path, 'r', encoding='utf-8')

        header = DictionaryHeader(SYSTEM_DICT_VERSION_2, int(time.time()), 'test')
        out_stream.write(header.to_bytes())
        builder = DictionaryBuilder(logger=self.logger)
        builder.build(lexicon_paths, matrix_input_stream, out_stream)
        out_stream.close()
        matrix_input_stream.close()

        buffers, header, grammar, lexicon_set = self.read_system_dictionary(out_path)
        lexicon = lexicon_set.lexicons[0]

        # header
        self.assertEqual(SYSTEM_DICT_VERSION_2, header.version)
        self.assertEqual('test', header.description)

        # grammar
        self.assertEqual(2, grammar.get_part_of_speech_size())
        self.assertEqual(["名詞", "固有名詞", "地名", "一般", "*", "*"], grammar.get_part_of_speech_string(0))
        self.assertEqual(["名詞", "普通名詞", "一般", "*", "*", "*"], grammar.get_part_of_speech_string(1))
        self.assertEqual(200, grammar.get_connect_cost(0, 0))

        # lexicon
        self.assertEqual(3, lexicon.size())
        self.assertEqual(0, lexicon.get_cost(0))
        wi = lexicon.get_word_info(0)
        self.assertEqual('東京都', wi.surface)
        self.assertEqual('東京都', wi.normalized_form)
        self.assertEqual(-1, wi.dictionary_form_word_id)
        self.assertEqual('ヒガシキョウト', wi.reading_form)
        self.assertEqual(0, wi.pos_id)
        self.assertEqual([1, 2], wi.a_unit_split)
        self.assertEqual([], wi.b_unit_split)
        self.assertEqual([1, 2], wi.synonym_group_ids)
        lst = lexicon.lookup('東京都'.encode('utf-8'), 0)
        self.assertEqual((0, len('東京都'.encode('utf-8'))), lst.__next__())
        with self.assertRaises(StopIteration):
            lst.__next__()

        self.assertEqual(-1, lexicon.get_left_id(1))
        self.assertEqual(0, lexicon.get_cost(1))
        wi = lexicon.get_word_info(1)
        self.assertEqual('東', wi.surface)
        self.assertEqual('ひがし', wi.normalized_form)
        self.assertEqual(-1, wi.dictionary_form_word_id)
        self.assertEqual('ヒガシ', wi.reading_form)
        self.assertEqual(1, wi.pos_id)
        self.assertEqual([], wi.a_unit_split)
        self.assertEqual([], wi.b_unit_split)
        lst = lexicon.lookup('東'.encode('utf-8'), 0)
        with self.assertRaises(StopIteration):
            lst.__next__()

    @staticmethod
    def read_system_dictionary(filename):
        """
        Copy of sudachipy.dictionary.Dictionary.read_system_dictionary
        :param filename:
        :return:
        """
        import mmap
        from sudachipy import dictionarylib
        buffers = []
        if filename is None:
            raise AttributeError("system dictionary is not specified")
        with open(filename, 'rb') as system_dic:
            bytes_ = mmap.mmap(system_dic.fileno(), 0, access=mmap.ACCESS_READ)
        buffers.append(bytes_)

        offset = 0
        header = dictionarylib.dictionaryheader.DictionaryHeader.from_bytes(bytes_, offset)
        if not header.is_system_dictionary():
            raise Exception("invalid system dictionary")
        offset += header.storage_size()

        grammar = dictionarylib.grammar.Grammar(bytes_, offset)
        offset += grammar.get_storage_size()

        lexicon = dictionarylib.lexiconset.LexiconSet(dictionarylib.doublearraylexicon.DoubleArrayLexicon(bytes_, offset, header.version == SYSTEM_DICT_VERSION_2))
        return buffers, header, grammar, lexicon



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

from sudachipy.plugin.input_text.default_input_text_plugin import DefaultInputTextPlugin
from sudachipy.utf8inputtextbuilder import UTF8InputTextBuilder

from tests import mock_grammar


class TestDefaultInputTextPlugin(unittest.TestCase):

    original_text = "ÂＢΓД㈱ｶﾞウ゛⼼Ⅲ"
    normalized_text = "âbγд(株)ガヴ⼼ⅲ"

    def setUp(self):
        self.builder = UTF8InputTextBuilder(self.original_text, mock_grammar.mocked_grammar)

        self.plugin = DefaultInputTextPlugin()

        try:
            self.plugin.set_up()
        except IOError:
            self.fail('no file')

        self.test_resources_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            os.pardir,
            'resources')

    def test_before_rewrite(self):
        self.assertEqual(self.original_text, self.builder.get_original_text())
        self.assertEqual(self.original_text, self.builder.get_text())
        text = self.builder.build()
        self.assertEqual(self.original_text, text.get_original_text())
        self.assertEqual(self.original_text, text.get_text())
        bytes_ = text.get_byte_text()
        self.assertEqual(30, len(bytes_))
        expected = b'\xc3\x82\xef\xbc\xa2\xce\x93\xd0\x94\xe3\x88\xb1\xef\xbd\xb6\xef\xbe\x9e\xe3\x82\xa6\xe3\x82\x9b\xe2\xbc\xbc\xe2\x85\xa2'
        self.assertEqual(expected, bytes_)
        self.assertEqual(0, text.get_original_index(0))
        self.assertEqual(0, text.get_original_index(1))
        self.assertEqual(1, text.get_original_index(2))
        self.assertEqual(1, text.get_original_index(4))
        self.assertEqual(3, text.get_original_index(8))
        self.assertEqual(5, text.get_original_index(12))
        self.assertEqual(9, text.get_original_index(24))
        self.assertEqual(9, text.get_original_index(26))

    def test_after_rewrite(self):
        self.assertEqual(self.original_text, self.builder.get_original_text())
        self.assertEqual(self.original_text, self.builder.get_text())
        self.plugin.rewrite(self.builder)
        text = self.builder.build()
        self.assertEqual(self.original_text, text.get_original_text())
        self.assertEqual(self.normalized_text, text.get_text())
        bytes_ = text.get_byte_text()
        self.assertEqual(24, len(bytes_))
        expected = b'\xc3\xa2\x62\xce\xb3\xd0\xb4\x28\xe6\xa0\xaa\x29\xe3\x82\xac\xe3\x83\xb4\xe2\xbc\xbc\xe2\x85\xb2'
        self.assertEqual(expected, bytes_)
        self.assertEqual(0, text.get_original_index(0))
        self.assertEqual(0, text.get_original_index(1))
        self.assertEqual(1, text.get_original_index(2))
        self.assertEqual(2, text.get_original_index(3))
        self.assertEqual(4, text.get_original_index(7))
        self.assertEqual(5, text.get_original_index(8))
        self.assertEqual(5, text.get_original_index(11))
        self.assertEqual(7, text.get_original_index(15))
        self.assertEqual(7, text.get_original_index(17))

    # def test_setup_with_null(self):

    def test_invalid_format_ignorelist(self):
        plugin = DefaultInputTextPlugin()
        with self.assertRaises(RuntimeError) as cm:
            plugin.read_rewrite_lists(os.path.join(self.test_resources_dir, 'rewrite_error_ignorelist.def'))
        self.assertEqual('12 is not character at line 1', cm.exception.args[0])

    def test_invalid_format_replacelist(self):
        plugin = DefaultInputTextPlugin()
        with self.assertRaises(RuntimeError) as cm:
            plugin.read_rewrite_lists(os.path.join(self.test_resources_dir, 'rewrite_error_replacelist.def'))
        self.assertEqual('invalid format at line 1', cm.exception.args[0])

    def test_duplicated_lines_replacelist(self):
        plugin = DefaultInputTextPlugin()
        with self.assertRaises(RuntimeError) as cm:
            plugin.read_rewrite_lists(os.path.join(self.test_resources_dir, 'rewrite_error_dup.def'))
        self.assertEqual('12 is already defined at line 2', cm.exception.args[0])


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from sudachipy.plugin.input_text import ProlongedSoundMarkInputTextPlugin
from sudachipy.utf8inputtextbuilder import UTF8InputTextBuilder

from tests.mock_grammar import mocked_grammar


class TestProlongedSoundMarkInputTextPlugin(TestCase):

    def setUp(self) -> None:
        self.plugin = ProlongedSoundMarkInputTextPlugin(None)
        for psm in ['ー', '〜', '〰']:
            self.plugin._psm_set.add(ord(psm))

    def test_combine_continuous_prolonged_sound_mark(self):
        original = 'ゴーール'
        normalized = 'ゴール'
        builder = UTF8InputTextBuilder(original, mocked_grammar)
        self.plugin.rewrite(builder)
        text = builder.build()

        self.assertEqual(original, text.original_text)
        self.assertEqual(normalized, text.get_text())
        bytes_ = text.get_byte_text()
        self.assertEqual(9, len(bytes_))

        self.assertEqual(b'\xe3\x82\xb4\xe3\x83\xbc\xe3\x83\xab', bytes_)
        self.assertEqual(0, text.get_original_index(0))
        self.assertEqual(1, text.get_original_index(3))
        self.assertEqual(3, text.get_original_index(6))
        self.assertEqual(4, text.get_original_index(9))

    def test_combined_continuous_prolonged_sound_marks_at_end(self):
        original = 'スーパーー'
        normalized = 'スーパー'
        builder = UTF8InputTextBuilder(original, mocked_grammar)
        self.plugin.rewrite(builder)
        text = builder.build()

        self.assertEqual(original, text.original_text)
        self.assertEqual(normalized, text.get_text())
        bytes_ = text.get_byte_text()
        self.assertEqual(12, len(bytes_))

        self.assertEqual(b'\xe3\x82\xb9\xe3\x83\xbc\xe3\x83\x91\xe3\x83\xbc', bytes_)
        self.assertEqual(0, text.get_original_index(0))
        self.assertEqual(1, text.get_original_index(3))
        self.assertEqual(2, text.get_original_index(6))
        self.assertEqual(3, text.get_original_index(9))
        self.assertEqual(5, text.get_original_index(12))

    def test_combine_continuous_prolonged_sound_marks_multi_times(self):
        original = 'エーービーーーシーーーー'
        normalized = 'エービーシー'
        builder = UTF8InputTextBuilder(original, mocked_grammar)
        self.plugin.rewrite(builder)
        text = builder.build()

        self.assertEqual(original, text.original_text)
        self.assertEqual(normalized, text.get_text())
        bytes_ = text.get_byte_text()
        self.assertEqual(18, len(bytes_))

        self.assertEqual(b'\xe3\x82\xa8\xe3\x83\xbc\xe3\x83\x93\xe3\x83\xbc\xe3\x82\xb7\xe3\x83\xbc', bytes_)
        self.assertEqual(0, text.get_original_index(0))
        self.assertEqual(1, text.get_original_index(3))
        self.assertEqual(3, text.get_original_index(6))
        self.assertEqual(4, text.get_original_index(9))
        self.assertEqual(7, text.get_original_index(12))
        self.assertEqual(8, text.get_original_index(15))
        self.assertEqual(12, text.get_original_index(18))

    def test_combine_continuous_prolonged_sound_marks_multi_symbol_types(self):
        original = 'エーービ〜〜〜シ〰〰〰〰'
        normalized = 'エービーシー'
        builder = UTF8InputTextBuilder(original, mocked_grammar)
        self.plugin.rewrite(builder)
        text = builder.build()

        self.assertEqual(original, text.original_text)
        self.assertEqual(normalized, text.get_text())
        bytes_ = text.get_byte_text()
        self.assertEqual(18, len(bytes_))

        self.assertEqual(b'\xe3\x82\xa8\xe3\x83\xbc\xe3\x83\x93\xe3\x83\xbc\xe3\x82\xb7\xe3\x83\xbc', bytes_)
        self.assertEqual(0, text.get_original_index(0))
        self.assertEqual(1, text.get_original_index(3))
        self.assertEqual(3, text.get_original_index(6))
        self.assertEqual(4, text.get_original_index(9))
        self.assertEqual(7, text.get_original_index(12))
        self.assertEqual(8, text.get_original_index(15))
        self.assertEqual(12, text.get_original_index(18))

    def test_combine_continuous_prolonged_sound_marks_multi_mixed_symbol_types(self):
        original = 'エー〜ビ〜〰ーシ〰ー〰〜'
        normalized = 'エービーシー'
        builder = UTF8InputTextBuilder(original, mocked_grammar)
        self.plugin.rewrite(builder)
        text = builder.build()

        self.assertEqual(original, text.original_text)
        self.assertEqual(normalized, text.get_text())
        bytes_ = text.get_byte_text()
        self.assertEqual(18, len(bytes_))

        self.assertEqual(b'\xe3\x82\xa8\xe3\x83\xbc\xe3\x83\x93\xe3\x83\xbc\xe3\x82\xb7\xe3\x83\xbc', bytes_)
        self.assertEqual(0, text.get_original_index(0))
        self.assertEqual(1, text.get_original_index(3))
        self.assertEqual(3, text.get_original_index(6))
        self.assertEqual(4, text.get_original_index(9))
        self.assertEqual(7, text.get_original_index(12))
        self.assertEqual(8, text.get_original_index(15))
        self.assertEqual(12, text.get_original_index(18))



import json
import os
import shutil
import tempfile
import time
from logging import getLogger
from unittest import TestCase

from sudachipy.dictionary import Dictionary
from sudachipy.dictionarylib import SYSTEM_DICT_VERSION_2
from sudachipy.dictionarylib.dictionarybuilder import DictionaryBuilder
from sudachipy.dictionarylib.dictionaryheader import DictionaryHeader


class TestSwitchDictionary(TestCase):

    def setUp(self):
        self.logger = getLogger()
        self.logger.disabled = True

        self.temp_dir = tempfile.mkdtemp()
        self.resource_dir = os.path.join(self.temp_dir, 'resources')
        os.makedirs(self.resource_dir)

        test_resource_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'resources')
        self.char_def_path = os.path.join(self.resource_dir, 'char.def')
        shutil.copy(os.path.join(test_resource_dir, 'char.def'), self.char_def_path)

        self.sudachi_json_path = os.path.join(self.resource_dir, 'sudachi.json')
        shutil.copy(os.path.join(test_resource_dir, 'sudachi.json'), self.sudachi_json_path)
        self._rewrite_json(self.sudachi_json_path, 'userDict', [])

        self.matrix_path = os.path.join(self.resource_dir, 'matrix.txt')
        with open(self.matrix_path, 'w', encoding='utf-8') as wf:
            wf.write('1 1\n0 0 200\n')

        small_lexs = ["島,0,0,0,島,名詞,普通名詞,一般,*,*,*,シマ,島,*,A,*,*,*"]
        core_lexs = ["徳島本町,0,0,0,徳島本町,名詞,固有名詞,地名,一般,*,*,トクシマホンチョウ,徳島本町,*,A,*,*,*,*"]
        notcore_lexs = ["徳島堰,0,0,0,徳島堰,名詞,固有名詞,一般,*,*,*,トクシマセギ,徳島堰,*,A,*,*,*"]

        small_lines = small_lexs
        core_lines = small_lexs + core_lexs
        full_lines = small_lexs + core_lexs + notcore_lexs

        self.small_txt_path = os.path.join(self.resource_dir, 'small.csv')
        self.core_txt_path = os.path.join(self.resource_dir, 'core.csv')
        self.full_txt_path = os.path.join(self.resource_dir, 'full.csv')

        self.small_dic_path = self._build_dictionary(self.small_txt_path, small_lines, 'small.dic')
        self.core_dic_path = self._build_dictionary(self.core_txt_path, core_lines, 'core.dic')
        self.full_dic_path = self._build_dictionary(self.full_txt_path, full_lines, 'full.dic')

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    @staticmethod
    def _rewrite_json(json_file_path, k, v):
        with open(json_file_path, 'r') as f:
            obj = json.load(f)
        obj[k] = v
        with open(json_file_path, 'w') as f:
            json.dump(obj, f, ensure_ascii=False, indent=4)

    def _build_dictionary(self, input_txt_path, lex_lines, dictionary_name):
        with open(input_txt_path, 'w', encoding='utf-8') as wf:
            wf.write("\n".join(lex_lines))

        out_path = os.path.join(self.resource_dir, dictionary_name)
        out_stream = open(out_path, 'wb')
        lexicon_paths = [input_txt_path]
        matrix_input_stream = open(self.matrix_path, 'r', encoding='utf-8')

        header = DictionaryHeader(SYSTEM_DICT_VERSION_2, int(time.time()), 'test')
        out_stream.write(header.to_bytes())
        builder = DictionaryBuilder(logger=self.logger)
        builder.build(lexicon_paths, matrix_input_stream, out_stream)
        out_stream.close()
        matrix_input_stream.close()

        return out_path

    def test_switch_dictionary(self):
        self._rewrite_json(self.sudachi_json_path, 'systemDict', 'small.dic')  # relative path
        self.dict = Dictionary(config_path=self.sudachi_json_path, resource_dir=self.resource_dir)
        self.assertEqual(1, self.dict.lexicon.size())
        self._rewrite_json(self.sudachi_json_path, 'systemDict', self.small_dic_path)  # abstract path
        self.dict = Dictionary(config_path=self.sudachi_json_path, resource_dir=self.resource_dir)
        self.assertEqual(1, self.dict.lexicon.size())

        self._rewrite_json(self.sudachi_json_path, 'systemDict', 'core.dic')
        self.dict = Dictionary(config_path=self.sudachi_json_path, resource_dir=self.resource_dir)
        self.assertEqual(2, self.dict.lexicon.size())
        self._rewrite_json(self.sudachi_json_path, 'systemDict', self.core_dic_path)
        self.dict = Dictionary(config_path=self.sudachi_json_path, resource_dir=self.resource_dir)
        self.assertEqual(2, self.dict.lexicon.size())

        self._rewrite_json(self.sudachi_json_path, 'systemDict', 'full.dic')
        self.dict = Dictionary(config_path=self.sudachi_json_path, resource_dir=self.resource_dir)
        self.assertEqual(3, self.dict.lexicon.size())
        self._rewrite_json(self.sudachi_json_path, 'systemDict', self.full_dic_path)
        self.dict = Dictionary(config_path=self.sudachi_json_path, resource_dir=self.resource_dir)
        self.assertEqual(3, self.dict.lexicon.size())



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from sudachipy import config

from . import JoinKatakanaOovPlugin, JoinNumericPlugin, PathRewritePlugin


def get_path_rewrite_plugin(json_obj) -> PathRewritePlugin:
    # In the future, users can define plugin by themselves
    try:
        if json_obj['class'] == 'sudachipy.plugin.path_rewrite.JoinNumericPlugin':
            return JoinNumericPlugin(json_obj)
        if json_obj['class'] == 'sudachipy.plugin.path_rewrite.JoinKatakanaOovPlugin':
            return JoinKatakanaOovPlugin(json_obj)
        raise ValueError('{} is invalid PathRewritePlugin class'.format(json_obj['class']))
    except KeyError:
        raise ValueError('config file is invalid format')


def get_path_rewrite_plugins() -> List[PathRewritePlugin]:
    if 'pathRewritePlugin' not in config.settings:
        return []
    ps = []
    for obj in config.settings['pathRewritePlugin']:
        ps.append(get_path_rewrite_plugin(obj))
    return ps



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

from sudachipy import tokenizer
from sudachipy.dictionary import Dictionary


class TestDictionary(unittest.TestCase):

    def setUp(self):
        resource_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'resources')
        self.dict_ = Dictionary(os.path.join(resource_dir, 'sudachi.json'), resource_dir=resource_dir)

    def tearDown(self) -> None:
        self.dict_.close()

    def test_create(self):
        self.assertEqual(tokenizer.Tokenizer, type(self.dict_.create()))

    def test_get_part_of_speech_size(self):
        self.assertEqual(9, self.dict_.grammar.get_part_of_speech_size())

    def test_get_part_of_speech_string(self):
        pos = self.dict_.grammar.get_part_of_speech_string(0)
        self.assertIsNotNone(pos)
        self.assertEqual('助動詞', pos[0])

    # def test_creat_with_merging_settings

    # def test_creat_with_merging_null_ settings


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import List

from sudachipy.dictionarylib.grammar import Grammar
from sudachipy.latticenode import LatticeNode
from sudachipy.utf8inputtext import UTF8InputText


class OovProviderPlugin(ABC):

    @abstractmethod
    def set_up(self, grammar: Grammar) -> None:
        raise NotImplementedError

    @abstractmethod
    def provide_oov(self, input_text: UTF8InputText, offset: int, has_other_words: bool) -> List[LatticeNode]:
        raise NotImplementedError

    def get_oov(self, input_text: UTF8InputText, offset: int, has_other_words: bool) -> List[LatticeNode]:
        nodes = self.provide_oov(input_text, offset, has_other_words)
        for node in nodes:
            node.set_begin(offset)
            node.set_end(offset + node.get_word_info().length())
        return nodes

    @staticmethod
    def create_node() -> LatticeNode:
        node = LatticeNode()
        node.set_oov()
        return node



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mmap
import os
import shutil
import tempfile
import unittest

from sudachipy.dictionarylib.grammar import Grammar


class TestGrammar(unittest.TestCase):

    alloc_size = 4096

    def setUp(self):
        storage = bytearray()
        self.build_partofspeech(storage)
        self.build_connect_table(storage)
        self.test_dir = tempfile.mkdtemp()
        f = os.path.join(self.test_dir, 'test_file.txt')
        with open(f, 'wb') as wf:
            wf.write(bytes(storage))
        self.mmap = None
        with open(f, 'rb') as rf:
            self.mmap = mmap.mmap(rf.fileno(), 0, access=mmap.ACCESS_COPY)
        self.storage_size = self.mmap.size()
        offset = 0
        self.grammar = Grammar(self.mmap, offset)

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_storage_size(self):
        self.assertEqual(self.storage_size, self.grammar.storage_size)

    def test_get_partofspeech_string(self):
        self.assertEqual(6, len(self.grammar.get_part_of_speech_string(0)))
        self.assertEqual("BOS/EOS", self.grammar.get_part_of_speech_string(0)[0])
        self.assertEqual("*", self.grammar.get_part_of_speech_string(0)[5])

        self.assertEqual("一般", self.grammar.get_part_of_speech_string(1)[1])
        self.assertEqual("*", self.grammar.get_part_of_speech_string(1)[5])

        self.assertEqual("五段-サ行", self.grammar.get_part_of_speech_string(2)[4])
        self.assertEqual("終止形-一般", self.grammar.get_part_of_speech_string(2)[5])

    def test_get_connect_cost(self):
        self.assertEqual(0, self.grammar.get_connect_cost(0, 0))
        self.assertEqual(-100, self.grammar.get_connect_cost(2, 1))
        self.assertEqual(200, self.grammar.get_connect_cost(1, 2))

    def test_set_connect_cost(self):
        self.grammar.set_connect_cost(0, 0, 300)
        self.assertEqual(300, self.grammar.get_connect_cost(0, 0))

    def test_get_bos_parameters(self):
        self.assertEqual(0, self.grammar.get_bos_parameter()[0])
        self.assertEqual(0, self.grammar.get_bos_parameter()[1])
        self.assertEqual(0, self.grammar.get_bos_parameter()[2])

    def test_get_eos_parameters(self):
        self.assertEqual(0, self.grammar.get_eos_parameter()[0])
        self.assertEqual(0, self.grammar.get_eos_parameter()[1])
        self.assertEqual(0, self.grammar.get_eos_parameter()[2])

    def test_read_from_file(self):
        # Todo
        # after tidying up dictionary management
        pass

    @staticmethod
    def build_partofspeech(storage):
        storage.extend((3).to_bytes(2, byteorder='little', signed=True))  # number of part of speech

        storage.extend(b'\x07B\x00O\x00S\x00/\x00E\x00O\x00S\x00\x01*\x00\x01*\x00\x01*\x00\x01*\x00\x01*\x00')

        storage.extend(b'\x02')
        storage.extend('名刺'.encode('utf-16-le'))
        storage.extend(b'\x02')
        storage.extend('一般'.encode('utf-16-le'))
        storage.extend(b'\x01*\x00\x01*\x00\x01*\x00\x01*\x00')

        storage.extend(b'\x02')
        storage.extend('動詞'.encode('utf-16-le'))
        storage.extend(b'\x02')
        storage.extend('一般'.encode('utf-16-le'))
        storage.extend(b'\x01*\x00\x01*\x00\x05')
        storage.extend('五段-サ行'.encode('utf-16-le'))
        storage.extend(b'\x06')
        storage.extend('終止形-一般'.encode('utf-16-le'))

    @staticmethod
    def build_connect_table(storage):
        storage.extend((3).to_bytes(2, byteorder='little', signed=True))  # number of leftId
        storage.extend((3).to_bytes(2, byteorder='little', signed=True))  # number of rightId

        storage.extend((0).to_bytes(2, byteorder='little', signed=True))  # number of rightId
        storage.extend((-300).to_bytes(2, byteorder='little', signed=True))  # number of rightId
        storage.extend((3000).to_bytes(2, byteorder='little', signed=True))  # number of rightId

        storage.extend((300).to_bytes(2, byteorder='little', signed=True))  # number of rightId
        storage.extend((-500).to_bytes(2, byteorder='little', signed=True))  # number of rightId
        storage.extend((-100).to_bytes(2, byteorder='little', signed=True))  # number of frightId

        storage.extend((-3000).to_bytes(2, byteorder='little', signed=True))  # number of rightId
        storage.extend((200).to_bytes(2, byteorder='little', signed=True))  # number of rightId
        storage.extend((2000).to_bytes(2, byteorder='little', signed=True))  # number of rightId


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sudachipy.dictionarylib.categorytype import CategoryType
from sudachipy.latticenode import LatticeNode
from sudachipy.plugin.path_rewrite.path_rewrite_plugin import PathRewritePlugin
from sudachipy.utf8inputtext import UTF8InputText


class JoinKatakanaOovPlugin(PathRewritePlugin):

    def __init__(self, json_obj):
        self.__pos = json_obj['oovPOS']
        self._min_length = 1
        if 'minLength' in json_obj:
            self._min_length = json_obj['minLength']
        self.oov_pos_id = None

    def set_up(self, grammar):
        if not self.__pos:
            raise ValueError("oovPOS is undefined")
        self.oov_pos_id = grammar.get_part_of_speech_id(self.__pos)
        if self.oov_pos_id < 0:
            raise ValueError("oovPOS is invalid")

    def rewrite(self, text, path, lattice):
        i = 0
        while True:
            if i >= len(path):
                break
            node = path[i]
            if not (node.is_oov() or self.is_shorter(self._min_length, text, node)) or \
                    not self.is_katakana_node(text, node):
                i += 1
                continue
            begin = i - 1
            while True:
                if begin < 0:
                    break
                if not self.is_katakana_node(text, path[begin]):
                    begin += 1
                    break
                begin -= 1
            begin = max(0, begin)
            end = i + 1
            while True:
                if end >= len(path):
                    break
                if not self.is_katakana_node(text, path[end]):
                    break
                end += 1
            pass
            while begin != end and not self.can_oov_bow_node(text, path[begin]):
                begin += 1
            if (end - begin) > 1:
                self.concatenate_oov(path, begin, end, self.oov_pos_id, lattice)
                i = begin + 1  # skip next node, as we already know it is not a joinable katakana
            i += 1

    def is_katakana_node(self, text, node):
        return CategoryType.KATAKANA in self.get_char_category_types(text, node)

    def is_one_char(self, text, node):
        b = node.get_begin()
        return b + text.get_code_points_offset_length(b, 1) == node.get_end()

    def can_oov_bow_node(self, text, node):
        return CategoryType.NOOOVBOW not in text.get_char_category_types(node.get_begin())

    @staticmethod
    def is_shorter(length: int, text: UTF8InputText, node: LatticeNode):
        return text.code_point_count(node.get_begin(), node.get_end()) < length



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from unittest import mock

from sudachipy.utf8inputtext import UTF8InputText


mocked_input_text = mock.Mock(spec=UTF8InputText)
text = ''
types = []


def set_text(text_: str) -> None:
    global text, types
    text = text_
    types = [set() for _ in text]


def set_category_type(begin: int, end: int, type_) -> None:
    global types
    for i in range(begin, end):
        types[i].add(type_)


mocked_input_text.get_text.return_value = text

mocked_input_text.get_original_text.return_value = text


def _mocked_get_substring(begin: int, end: int) -> str:
    global text
    return text[begin:end]


mocked_input_text.get_substring.side_effect = _mocked_get_substring


def _mocked_get_char_category_types(begin: int, end: int = None) -> set:
    global text, types
    if end is None:
        return types[begin]
    continuous_category = deepcopy(types[begin])
    for i in range(begin + 1, end):
        continuous_category = continuous_category.intersection(types[i])
    return continuous_category


mocked_input_text.get_char_category_types.side_effect = _mocked_get_char_category_types


def _mocked_get_char_category_continuous_length(idx: int) -> int:
    global text, types
    continuous_category = deepcopy(types[idx])
    for i in range(idx + 1, len(text)):
        continuous_category = continuous_category.intersection(types[i])
        if not continuous_category:
            return i - idx
    return len(text) - idx


mocked_input_text.get_char_category_continuous_length.side_effect = _mocked_get_char_category_continuous_length


def _mocked_get_code_points_offset_length(idx: int, offset: int) -> int:
    return offset


mocked_input_text.get_code_points_offset_length.side_effect = _mocked_get_code_points_offset_length



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

import sudachipy
import sudachipy.dictionarylib as dictionarylib


class TestUTF8InputText(unittest.TestCase):

    def setUp(self):
        self.TEXT = "âｂC1あ234漢字𡈽アｺﾞ"
        self.bytes = [
            b'0xC3', b'0xA2', b'0xEF', b'0xBD', b'0x82',
            b'0x43', b'0x31', b'0xE3', b'0x81', b'0x82',
            b'0x32', b'0x33', b'0x34', b'0xE6', b'0xBC',
            b'0xA2', b'0xE5', b'0xAD', b'0x97', b'0xF0',
            b'0xA1', b'0x88', b'0xBD', b'0xE3', b'0x82',
            b'0xA2', b'0xEF', b'0xBD', b'0xBA', b'0xEF',
            b'0xBE', b'0x9E'
        ]

        self.input = None

        grammar = self.MockGrammar()
        char_category = dictionarylib.charactercategory.CharacterCategory()
        this_dir = os.path.dirname(os.path.abspath(__file__))
        char_category.read_character_definition(os.path.join(this_dir, 'resources/char.def'))
        grammar.set_character_category(char_category)

        self.builder = sudachipy.utf8inputtextbuilder.UTF8InputTextBuilder(self.TEXT, grammar)

    def test_get_original_text(self):
        self.assertEqual(self.builder.get_original_text(), self.TEXT)
        self.assertEqual(self.builder.get_text(), self.TEXT)
        self.input = self.builder.build()
        self.assertEqual(self.input.get_original_text(), self.TEXT)
        self.assertEqual(self.input.get_text(), self.TEXT)

    def test_get_byte_text(self):
        input_ = self.builder.build()
        self.assertEqual(len(input_.get_byte_text()), 32)
        self.assertEqual(self.TEXT.encode('utf-8'), input_.get_byte_text())

    def test_get_original_index(self):
        input_ = self.builder.build()
        self.assertEqual(input_.get_original_index(0), 0)
        self.assertEqual(input_.get_original_index(1), 0)
        self.assertEqual(input_.get_original_index(2), 1)
        self.assertEqual(input_.get_original_index(4), 1)
        self.assertEqual(input_.get_original_index(6), 3)
        self.assertEqual(input_.get_original_index(7), 4)
        self.assertEqual(input_.get_original_index(10), 5)
        self.assertEqual(input_.get_original_index(18), 9)
        self.assertEqual(input_.get_original_index(19), 10)
        self.assertEqual(input_.get_original_index(22), 10)
        self.assertEqual(input_.get_original_index(23), 11)
        self.assertEqual(input_.get_original_index(28), 12)
        self.assertEqual(input_.get_original_index(31), 13)

    def test_get_char_category_types(self):
        input_ = self.builder.build()
        self.assertTrue(dictionarylib.categorytype.CategoryType.ALPHA in input_.get_char_category_types(0))
        self.assertTrue(dictionarylib.categorytype.CategoryType.ALPHA in input_.get_char_category_types(2))
        self.assertTrue(dictionarylib.categorytype.CategoryType.ALPHA in input_.get_char_category_types(5))
        self.assertTrue(dictionarylib.categorytype.CategoryType.NUMERIC in input_.get_char_category_types(6))
        self.assertTrue(dictionarylib.categorytype.CategoryType.HIRAGANA in input_.get_char_category_types(7))
        self.assertTrue(dictionarylib.categorytype.CategoryType.HIRAGANA in input_.get_char_category_types(9))
        self.assertTrue(dictionarylib.categorytype.CategoryType.NUMERIC in input_.get_char_category_types(10))
        self.assertTrue(dictionarylib.categorytype.CategoryType.KANJI in input_.get_char_category_types(13))
        self.assertTrue(dictionarylib.categorytype.CategoryType.KANJI in input_.get_char_category_types(18))
        self.assertTrue(dictionarylib.categorytype.CategoryType.DEFAULT in input_.get_char_category_types(19))
        self.assertTrue(dictionarylib.categorytype.CategoryType.DEFAULT in input_.get_char_category_types(22))
        self.assertTrue(dictionarylib.categorytype.CategoryType.KATAKANA in input_.get_char_category_types(23))
        self.assertTrue(dictionarylib.categorytype.CategoryType.KATAKANA in input_.get_char_category_types(26))
        self.assertTrue(dictionarylib.categorytype.CategoryType.KATAKANA in input_.get_char_category_types(31))

    def test_get_char_category_continuous_length(self):
        input_ = self.builder.build()
        self.assertEqual(input_.get_char_category_continuous_length(0), 6)
        self.assertEqual(input_.get_char_category_continuous_length(1), 5)
        self.assertEqual(input_.get_char_category_continuous_length(2), 4)
        self.assertEqual(input_.get_char_category_continuous_length(5), 1)
        self.assertEqual(input_.get_char_category_continuous_length(6), 1)
        self.assertEqual(input_.get_char_category_continuous_length(7), 3)
        self.assertEqual(input_.get_char_category_continuous_length(10), 3)
        self.assertEqual(input_.get_char_category_continuous_length(11), 2)
        self.assertEqual(input_.get_char_category_continuous_length(12), 1)
        self.assertEqual(input_.get_char_category_continuous_length(19), 4)
        self.assertEqual(input_.get_char_category_continuous_length(22), 1)
        self.assertEqual(input_.get_char_category_continuous_length(23), 9)
        self.assertEqual(input_.get_char_category_continuous_length(26), 6)
        self.assertEqual(input_.get_char_category_continuous_length(31), 1)

    def test_replace_with_same_length(self):
        self.builder.replace(8, 10, "ああ")
        self.assertEqual(self.builder.get_original_text(), self.TEXT)
        self.assertEqual(self.builder.get_text(), "âｂC1あ234ああ𡈽アｺﾞ")
        input_ = self.builder.build()
        self.assertEqual(input_.get_original_text(), self.TEXT)
        self.assertEqual(input_.get_text(), "âｂC1あ234ああ𡈽アｺﾞ")
        self.assertEqual(len(input_.get_byte_text()), 32)
        self.assertEqual(input_.get_original_index(0), 0)
        self.assertEqual(input_.get_original_index(12), 7)
        self.assertEqual(input_.get_original_index(13), 8)
        self.assertEqual(input_.get_original_index(15), 8)
        self.assertEqual(input_.get_original_index(16), 10)
        self.assertEqual(input_.get_original_index(18), 10)
        self.assertEqual(input_.get_original_index(19), 10)
        self.assertEqual(input_.get_original_index(22), 10)
        self.assertEqual(input_.get_original_index(31), 13)

    def test_replaceWithDeletion(self):
        self.builder.replace(8, 10, "あ")
        self.assertEqual(self.builder.get_original_text(), self.TEXT)
        self.assertEqual(self.builder.get_text(), "âｂC1あ234あ𡈽アｺﾞ")
        input_ = self.builder.build()
        self.assertEqual(input_.get_original_text(), self.TEXT)
        self.assertEqual(input_.get_text(), "âｂC1あ234あ𡈽アｺﾞ")
        self.assertEqual(len(input_.get_byte_text()), 29)
        self.assertEqual(input_.get_original_index(0), 0)
        self.assertEqual(input_.get_original_index(12), 7)
        self.assertEqual(input_.get_original_index(13), 8)
        self.assertEqual(input_.get_original_index(15), 8)
        self.assertEqual(input_.get_original_index(16), 10)
        self.assertEqual(input_.get_original_index(19), 10)
        self.assertEqual(input_.get_original_index(28), 13)

    def test_replaceWithInsertion(self):
        self.builder.replace(8, 10, "あああ")
        self.assertEqual(self.builder.get_original_text(), self.TEXT)
        self.assertEqual(self.builder.get_text(), "âｂC1あ234あああ𡈽アｺﾞ")
        input_ = self.builder.build()
        self.assertEqual(input_.get_original_text(), self.TEXT)
        self.assertEqual(input_.get_text(), "âｂC1あ234あああ𡈽アｺﾞ")
        self.assertEqual(len(input_.get_byte_text()), 35)
        self.assertEqual(input_.get_original_index(0), 0)  # â
        self.assertEqual(input_.get_original_index(12), 7)  # 4
        self.assertEqual(input_.get_original_index(13), 8)  # >あ< ああ
        self.assertEqual(input_.get_original_index(21), 10)  # ああ >あ<
        self.assertEqual(input_.get_original_index(22), 10)  # 𡈽
        self.assertEqual(input_.get_original_index(25), 10)  # 𡈽
        self.assertEqual(input_.get_original_index(35), 14)  #  ﾞ

    def test_replaceMultiTimes(self):
        self.builder.replace(0, 1, "a")
        self.builder.replace(1, 2, "b")
        self.builder.replace(2, 3, "c")
        self.builder.replace(10, 11, "土")
        self.builder.replace(12, 14, "ゴ")
        input_ = self.builder.build()
        self.assertEqual(input_.get_original_text(), self.TEXT)
        self.assertEqual(input_.get_text(), "abc1あ234漢字土アゴ")
        self.assertEqual(len(input_.get_byte_text()), 25)
        self.assertEqual(input_.get_original_index(0), 0)
        self.assertEqual(input_.get_original_index(1), 1)
        self.assertEqual(input_.get_original_index(2), 2)
        self.assertEqual(input_.get_original_index(7), 5)
        self.assertEqual(input_.get_original_index(8), 6)
        self.assertEqual(input_.get_original_index(9), 7)
        self.assertEqual(input_.get_original_index(15), 9)
        self.assertEqual(input_.get_original_index(16), 10)
        self.assertEqual(input_.get_original_index(18), 10)
        self.assertEqual(input_.get_original_index(19), 11)
        self.assertEqual(input_.get_original_index(21), 11)
        self.assertEqual(input_.get_original_index(22), 12)
        self.assertEqual(input_.get_original_index(24), 12)

    def test_getByteLengthByCodePoints(self):
        input_ = self.builder.build()
        self.assertEqual(input_.get_code_points_offset_length(0, 1), 2)
        self.assertEqual(input_.get_code_points_offset_length(0, 4), 7)
        self.assertEqual(input_.get_code_points_offset_length(10, 1), 1)
        self.assertEqual(input_.get_code_points_offset_length(11, 1), 1)
        self.assertEqual(input_.get_code_points_offset_length(12, 1), 1)
        self.assertEqual(input_.get_code_points_offset_length(13, 2), 6)
        self.assertEqual(input_.get_code_points_offset_length(19, 1), 4)
        self.assertEqual(input_.get_code_points_offset_length(23, 3), 9)

    def test_codePointCount(self):
        input_ = self.builder.build()
        self.assertEqual(input_.code_point_count(0, 2), 1)
        self.assertEqual(input_.code_point_count(0, 7), 4)
        self.assertEqual(input_.code_point_count(13, 19), 2)

    def test_canBow(self):
        input_ = self.builder.build()
        self.assertTrue(input_.can_bow(0))  # â
        self.assertFalse(input_.can_bow(1))
        self.assertFalse(input_.can_bow(2))  # ｂ
        self.assertFalse(input_.can_bow(3))
        self.assertFalse(input_.can_bow(4))
        self.assertFalse(input_.can_bow(5))  # C
        self.assertTrue(input_.can_bow(6))  # 1
        self.assertTrue(input_.can_bow(7))  # あ

        self.assertTrue(input_.can_bow(19))  # 𡈽
        self.assertFalse(input_.can_bow(20))
        self.assertFalse(input_.can_bow(21))
        self.assertFalse(input_.can_bow(22))
        self.assertTrue(input_.can_bow(23))  # ア

    def test_getWordCandidateLength(self):
        input_ = self.builder.build()
        self.assertEqual(input_.get_word_candidate_length(0), 6)
        self.assertEqual(input_.get_word_candidate_length(6), 1)
        self.assertEqual(input_.get_word_candidate_length(19), 4)
        self.assertEqual(input_.get_word_candidate_length(29), 3)

    class MockGrammar:
        char_category = None

        def get_part_of_speech_size(self):
            return 0

        def get_part_of_speech_string(self, pos_id):
            return None

        def get_part_of_speech_id(self, pos):
            return 0

        def get_connect_cost(self, left_id, right_id, cost=None):
            if cost is None:
                return 0
            else:
                return

        def get_bos_parameter(self):
            return None

        def get_eos_parameter(self):
            return None

        def get_character_category(self):
            return self.char_category

        def set_character_category(self, char_category):
            self.char_category = char_category


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

from sudachipy.dictionary import Dictionary
from sudachipy.plugin.path_rewrite import JoinNumericPlugin
from sudachipy.utf8inputtextbuilder import UTF8InputTextBuilder


class TestJoinNumericOOVPlugin(unittest.TestCase):

    def setUp(self):
        pass
        resource_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, 'resources')
        self.dict_ = Dictionary(os.path.join(resource_dir, 'numeric_sudachi.json'), resource_dir)
        self.tokenizer = self.dict_.create()
        self.plugin = JoinNumericPlugin(None)
        self.plugin.set_up(self.dict_.grammar)

    def test_digit(self):
        path = self.get_path('123円20銭')
        self.assertEqual(4, len(path))
        self.assertEqual('123', path[0].get_word_info().surface)
        self.assertEqual('20', path[2].get_word_info().surface)

        path = self.get_path('080-121')
        self.assertEqual(3, len(path))
        self.assertEqual('080', path[0].get_word_info().surface)
        self.assertEqual('121', path[2].get_word_info().surface)

    def test_kanji_numeric(self):
        path = self.get_path('一二三万二千円')
        self.assertEqual(2, len(path))
        self.assertEqual('一二三万二千', path[0].get_word_info().surface)

        path = self.get_path('二百百')
        self.assertEqual(3, len(path))

    def test_normalize(self):
        self.plugin._enable_normalize = True
        path = self.get_path('一二三万二千円')
        self.assertEqual(2, len(path))
        self.assertEqual('1232000', path[0].get_word_info().normalized_form)

    def test_normalized_with_not_numeric(self):
        self.plugin._enable_normalize = True
        path = self.get_path('六三四')
        self.assertEqual(1, len(path))
        self.assertEqual('六三四', path[0].get_word_info().normalized_form)

    def test_point(self):
        self.plugin._enable_normalize = True

        path = self.get_path('1.002')
        self.assertEqual(1, len(path))
        self.assertEqual('1.002', path[0].get_word_info().normalized_form)

        path = self.get_path('.002')
        self.assertEqual(2, len(path))
        self.assertEqual('.', path[0].get_word_info().normalized_form)
        self.assertEqual('002', path[1].get_word_info().normalized_form)

        path = self.get_path('22.')
        self.assertEqual(2, len(path))
        self.assertEqual('22', path[0].get_word_info().normalized_form)
        self.assertEqual('.', path[1].get_word_info().normalized_form)

        path = self.get_path('22.節')
        self.assertEqual(3, len(path))
        self.assertEqual('22', path[0].get_word_info().normalized_form)
        self.assertEqual('.', path[1].get_word_info().normalized_form)

        path = self.get_path('.c')
        self.assertEqual(2, len(path))
        self.assertEqual('.', path[0].get_word_info().normalized_form)

        path = self.get_path('1.20.3')
        self.assertEqual(5, len(path))
        self.assertEqual('20', path[2].get_word_info().normalized_form)

        path = self.get_path('652...')
        self.assertEqual(4, len(path))
        self.assertEqual('652', path[0].get_word_info().normalized_form)

    def test_comma(self):
        self.plugin._enable_normalize = True

        path = self.get_path('2,000,000')
        self.assertEqual(1, len(path))
        self.assertEqual('2000000', path[0].get_word_info().normalized_form)

        path = self.get_path('2,00,000,000円')
        self.assertEqual(8, len(path))
        self.assertEqual('2', path[0].get_word_info().normalized_form)
        self.assertEqual(',', path[1].get_word_info().normalized_form)
        self.assertEqual('00', path[2].get_word_info().normalized_form)
        self.assertEqual(',', path[3].get_word_info().normalized_form)
        self.assertEqual('000', path[4].get_word_info().normalized_form)
        self.assertEqual(',', path[5].get_word_info().normalized_form)
        self.assertEqual('000', path[6].get_word_info().normalized_form)

        path = self.get_path(',')
        self.assertEqual(1, len(path))

        path = self.get_path('652,,,')
        self.assertEqual(4, len(path))
        self.assertEqual('652', path[0].get_word_info().normalized_form)

        path = self.get_path('256,5.50389')
        self.assertEqual(3, len(path))
        self.assertEqual('256', path[0].get_word_info().normalized_form)
        self.assertEqual('5.50389', path[2].get_word_info().normalized_form)

        path = self.get_path('256,550.389')
        self.assertEqual(1, len(path))
        self.assertEqual('256550.389', path[0].get_word_info().normalized_form)

    def test_single_node(self):
        self.plugin._enable_normalize = False
        path = self.get_path('猫三匹')
        self.assertEqual(3, len(path))
        self.assertEqual('三', path[1].get_word_info().normalized_form)

        self.plugin._enable_normalize = True
        path = self.get_path('猫三匹')
        self.assertEqual(3, len(path))
        self.assertEqual('3', path[1].get_word_info().normalized_form)

    def get_path(self, text: str):
        input_ = UTF8InputTextBuilder(text, self.tokenizer._grammar).build()
        self.tokenizer._build_lattice(input_)
        path = self.tokenizer._lattice.get_best_path()
        self.plugin.rewrite(input_, path, self.tokenizer._lattice)
        self.tokenizer._lattice.clear()
        return path


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# the first version of system dictionaries
SYSTEM_DICT_VERSION_1 = 0x7366d3f18bd111e7

# the second version of system dictionaries
SYSTEM_DICT_VERSION_2 = 0xce9f011a92394434

# the first version of user dictionaries
USER_DICT_VERSION_1 = 0xa50f31188bd211e7

# the second version of user dictionaries
USER_DICT_VERSION_2 = 0x9fdeb5a90168d868

# the third version of user dictionaries
USER_DICT_VERSION_3 = 0xca9811756ff64fb0


def is_dictionary(version):
    return version in [
        SYSTEM_DICT_VERSION_1, SYSTEM_DICT_VERSION_2,
        USER_DICT_VERSION_1, USER_DICT_VERSION_2, USER_DICT_VERSION_3
    ]


def is_system_dictionary(version):
    return version == SYSTEM_DICT_VERSION_1 or version == SYSTEM_DICT_VERSION_2


def is_user_dictionary(version):
    return version == USER_DICT_VERSION_1 or version == USER_DICT_VERSION_2 or version == USER_DICT_VERSION_3


def has_grammar(version):
    return is_system_dictionary(version) or version == USER_DICT_VERSION_2 or version == USER_DICT_VERSION_3


def has_synonym_group_ids(version):
    return version == SYSTEM_DICT_VERSION_2 or version == USER_DICT_VERSION_3



from .latticenode cimport LatticeNode

cdef extern from "limits.h":
    cdef int INT_MAX

cdef class Lattice:

    cdef int size
    cdef int capacity
    cdef LatticeNode eos_node

    cdef list end_lists
    cdef object grammar
    cdef object eos_params
    cdef const short[:,:] connect_costs

    cpdef void resize(self, int size)
    cpdef void insert(self, int begin, int end, LatticeNode node)
    cdef void connect_node(self, LatticeNode r_node)
    cdef void connect_eos_node(self)



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct


class WordIdTable(object):
    def __init__(self, bytes_, offset):
        bytes_.seek(offset)
        self.size = int.from_bytes(bytes_.read(4), 'little')
        self.offset = offset + 4
        self._bytes_view = memoryview(bytes_)[self.offset: self.offset + self.size]

    def __del__(self):
        self._bytes_view.release()

    def storage_size(self):
        return 4 + self.size

    def get(self, index):
        length = self._bytes_view[index]
        result = struct.unpack_from("<{}I".format(length), self._bytes_view, index + 1)
        return result



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class WordInfo:
    def __init__(self,
                 surface,
                 head_word_length,
                 pos_id,
                 normalized_form,
                 dictionary_form_word_id,
                 dictionary_form,
                 reading_form,
                 a_unit_split,
                 b_unit_split,
                 word_structure,
                 synonym_group_ids):
        self.surface = surface
        self.head_word_length = head_word_length
        self.pos_id = pos_id
        self.normalized_form = normalized_form
        self.dictionary_form_word_id = dictionary_form_word_id
        self.dictionary_form = dictionary_form
        self.reading_form = reading_form
        self.a_unit_split = a_unit_split
        self.b_unit_split = b_unit_split
        self.word_structure = word_structure
        self.synonym_group_ids = synonym_group_ids

    def length(self):
        return self.head_word_length
